# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

import dearpygui.dearpygui as dpg
from deareis.utility import calculate_window_position_dimensions
from deareis.tooltips import attach_tooltip
import deareis.tooltips as tooltips
from deareis.enums import (
    Method,
    Mode,
    Test,
    Weight,
    label_to_method,
    label_to_mode,
    label_to_test,
    label_to_weight,
    method_to_label,
    mode_to_label,
    test_to_label,
    weight_to_label,
)
from deareis.data import (
    FitSettings,
    SimulationSettings,
    TestSettings,
)
from deareis.config import (
    DEFAULT_TEST_SETTINGS,
    DEFAULT_FIT_SETTINGS,
    DEFAULT_SIMULATION_SETTINGS,
)
from deareis.gui.plotting.export import (
    EXTENSIONS,
    LEGEND_LOCATIONS,
    PREVIEW_LIMITS,
    UNITS_PER_INCH,
)
from deareis.signals import Signal
import deareis.signals as signals


def show_defaults_settings_window(state):
    x: int
    y: int
    w: int
    h: int
    x, y, w, h = calculate_window_position_dimensions(374, 540)

    window: int = dpg.generate_uuid()
    key_handler: int = dpg.generate_uuid()

    def close_window():
        if dpg.does_item_exist(window):
            dpg.delete_item(window)
        if dpg.does_item_exist(key_handler):
            dpg.delete_item(key_handler)
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)

    with dpg.handler_registry(tag=key_handler):
        dpg.add_key_release_handler(
            key=dpg.mvKey_Escape,
            callback=close_window,
        )

    with dpg.window(
        label="Settings - defaults",
        modal=True,
        pos=(
            x,
            y,
        ),
        width=w,
        height=h,
        no_resize=True,
        on_close=close_window,
        tag=window,
    ):
        label_pad: int
        # TODO: Timeout when generating certain outputs (e.g. SymPy and LaTeX expressions)
        # with dpg.collapsing_header(label="General", default_open=True):
        with dpg.collapsing_header(label="Kramers-Kronig tab", default_open=True):
            test_test_combo: int = dpg.generate_uuid()
            test_mode_combo: int = dpg.generate_uuid()
            test_mu_crit_slider: int = dpg.generate_uuid()
            test_add_cap_checkbox: int = dpg.generate_uuid()
            test_add_ind_checkbox: int = dpg.generate_uuid()
            test_method_combo: int = dpg.generate_uuid()
            test_max_nfev_input: int = dpg.generate_uuid()

            def update_default_test_settings():
                state.config.default_test_settings = TestSettings(
                    label_to_test.get(dpg.get_value(test_test_combo), Test.COMPLEX),
                    label_to_mode.get(dpg.get_value(test_mode_combo), Mode.EXPLORATORY),
                    999,
                    dpg.get_value(test_mu_crit_slider),
                    dpg.get_value(test_add_cap_checkbox),
                    dpg.get_value(test_add_ind_checkbox),
                    label_to_method.get(
                        dpg.get_value(test_method_combo), Method.LEASTSQ
                    ),
                    dpg.get_value(test_max_nfev_input),
                )

            def restore_default_test_settings():
                dpg.set_value(
                    test_test_combo, test_to_label[DEFAULT_TEST_SETTINGS.test]
                )
                dpg.set_value(
                    test_mode_combo, mode_to_label[DEFAULT_TEST_SETTINGS.mode]
                )
                dpg.set_value(test_mu_crit_slider, DEFAULT_TEST_SETTINGS.mu_criterion)
                dpg.set_value(
                    test_add_cap_checkbox, DEFAULT_TEST_SETTINGS.add_capacitance
                )
                dpg.set_value(
                    test_add_ind_checkbox, DEFAULT_TEST_SETTINGS.add_inductance
                )
                dpg.set_value(
                    test_method_combo, method_to_label[DEFAULT_TEST_SETTINGS.method]
                )
                dpg.set_value(test_max_nfev_input, DEFAULT_TEST_SETTINGS.max_nfev)
                update_default_test_settings()

            label_pad = 22
            with dpg.child_window(border=False, width=350, height=180):
                with dpg.group(horizontal=True):
                    dpg.add_text("Test".rjust(label_pad))
                    attach_tooltip(tooltips.kramers_kronig.test)
                    dpg.add_combo(
                        items=list(label_to_test.keys()),
                        default_value=test_to_label.get(
                            state.config.default_test_settings.test
                        ),
                        callback=update_default_test_settings,
                        tag=test_test_combo,
                        width=-1,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Mode".rjust(label_pad))
                    attach_tooltip(tooltips.kramers_kronig.mode)
                    dpg.add_combo(
                        items=list(label_to_mode.keys()),
                        default_value=mode_to_label.get(
                            state.config.default_test_settings.mode
                        ),
                        callback=update_default_test_settings,
                        tag=test_mode_combo,
                        width=-1,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Add capacitor".rjust(label_pad))
                    attach_tooltip(tooltips.kramers_kronig.add_capacitance)
                    dpg.add_checkbox(
                        default_value=state.config.default_test_settings.add_capacitance,
                        callback=update_default_test_settings,
                        tag=test_add_cap_checkbox,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Add inductor".rjust(label_pad))
                    attach_tooltip(tooltips.kramers_kronig.add_inductance)
                    dpg.add_checkbox(
                        default_value=state.config.default_test_settings.add_inductance,
                        callback=update_default_test_settings,
                        tag=test_add_ind_checkbox,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Mu-criterion".rjust(label_pad))
                    attach_tooltip(tooltips.kramers_kronig.mu_criterion)
                    dpg.add_slider_float(
                        default_value=state.config.default_test_settings.mu_criterion,
                        min_value=0.01,
                        max_value=0.99,
                        clamped=True,
                        format="%.2f",
                        callback=update_default_test_settings,
                        tag=test_mu_crit_slider,
                        width=-1,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Fitting method".rjust(label_pad))
                    attach_tooltip(tooltips.kramers_kronig.method)
                    dpg.add_combo(
                        items=list(label_to_method.keys()),
                        default_value=method_to_label.get(
                            state.config.default_test_settings.method
                        ),
                        callback=update_default_test_settings,
                        tag=test_method_combo,
                        width=-1,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Max. num. func. eval.".rjust(label_pad))
                    attach_tooltip(tooltips.kramers_kronig.nfev)
                    dpg.add_input_int(
                        default_value=state.config.default_test_settings.max_nfev,
                        min_value=0,
                        min_clamped=True,
                        step=0,
                        on_enter=True,
                        callback=update_default_test_settings,
                        tag=test_max_nfev_input,
                        width=-1,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("".rjust(label_pad))
                    dpg.add_button(
                        label="Restore defaults",
                        callback=restore_default_test_settings,
                    )
            dpg.add_spacer(height=8)

        with dpg.collapsing_header(label="Fitting tab", default_open=True):
            fit_cdc_input: int = dpg.generate_uuid()
            fit_method_combo: int = dpg.generate_uuid()
            fit_weight_combo: int = dpg.generate_uuid()
            fit_max_nfev_input: int = dpg.generate_uuid()

            def update_default_fit_settings():
                state.config.default_fit_settings = FitSettings(
                    dpg.get_value(fit_cdc_input),
                    label_to_method.get(dpg.get_value(fit_method_combo), Method.AUTO),
                    label_to_weight.get(dpg.get_value(fit_weight_combo), Weight.AUTO),
                    dpg.get_value(fit_max_nfev_input),
                )

            def restore_default_fit_settings():
                dpg.set_value(fit_cdc_input, DEFAULT_FIT_SETTINGS.cdc)
                dpg.set_value(
                    fit_method_combo,
                    method_to_label.get(DEFAULT_FIT_SETTINGS.method, "Auto"),
                )
                dpg.set_value(
                    fit_weight_combo,
                    weight_to_label.get(DEFAULT_FIT_SETTINGS.weight, "Auto"),
                )
                dpg.set_value(fit_max_nfev_input, DEFAULT_FIT_SETTINGS.max_nfev)
                update_default_fit_settings()

            label_pad = 22
            with dpg.child_window(border=False, width=350, height=112):
                with dpg.group(horizontal=True):
                    dpg.add_text("Circuit".rjust(label_pad))
                    attach_tooltip(tooltips.fitting.cdc)
                    dpg.add_input_text(
                        default_value=state.config.default_fit_settings.cdc,
                        on_enter=True,
                        callback=update_default_fit_settings,
                        width=-1,
                        tag=fit_cdc_input,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Method".rjust(label_pad))
                    attach_tooltip(tooltips.fitting.method)
                    dpg.add_combo(
                        items=["Auto"] + list(label_to_method.keys()),
                        default_value=method_to_label.get(
                            state.config.default_fit_settings.method, "Auto"
                        ),
                        width=-1,
                        callback=update_default_fit_settings,
                        tag=fit_method_combo,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Weight function".rjust(label_pad))
                    attach_tooltip(tooltips.fitting.weight)
                    dpg.add_combo(
                        items=["Auto"] + list(label_to_weight.keys()),
                        default_value=weight_to_label.get(
                            state.config.default_fit_settings.weight, "Auto"
                        ),
                        width=-1,
                        callback=update_default_fit_settings,
                        tag=fit_weight_combo,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Max. num. func. eval.".rjust(label_pad))
                    attach_tooltip(tooltips.fitting.nfev)
                    dpg.add_input_int(
                        default_value=state.config.default_fit_settings.max_nfev,
                        min_value=0,
                        min_clamped=True,
                        step=0,
                        on_enter=True,
                        callback=update_default_fit_settings,
                        tag=fit_max_nfev_input,
                        width=-1,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("".rjust(label_pad))
                    dpg.add_button(
                        label="Restore defaults",
                        callback=restore_default_fit_settings,
                    )
            dpg.add_spacer(height=8)

        with dpg.collapsing_header(label="Simulation tab", default_open=True):
            sim_cdc_input: int = dpg.generate_uuid()
            sim_max_freq_input: int = dpg.generate_uuid()
            sim_min_freq_input: int = dpg.generate_uuid()
            sim_per_decade_input: int = dpg.generate_uuid()

            def update_default_sim_settings():
                state.config.default_simulation_settings = SimulationSettings(
                    dpg.get_value(sim_cdc_input),
                    dpg.get_value(sim_min_freq_input),
                    dpg.get_value(sim_max_freq_input),
                    dpg.get_value(sim_per_decade_input),
                )

            def restore_default_sim_settings():
                dpg.set_value(sim_cdc_input, DEFAULT_SIMULATION_SETTINGS.cdc)
                dpg.set_value(
                    sim_min_freq_input, DEFAULT_SIMULATION_SETTINGS.min_frequency
                )
                dpg.set_value(
                    sim_max_freq_input, DEFAULT_SIMULATION_SETTINGS.max_frequency
                )
                dpg.set_value(
                    sim_per_decade_input, DEFAULT_SIMULATION_SETTINGS.num_per_decade
                )
                update_default_sim_settings()

            label_pad = 22
            with dpg.child_window(border=False, width=350, height=112):
                with dpg.group(horizontal=True):
                    dpg.add_text("Circuit".rjust(label_pad))
                    attach_tooltip(tooltips.simulation.cdc)
                    dpg.add_input_text(
                        default_value=state.config.default_simulation_settings.cdc,
                        on_enter=True,
                        callback=update_default_sim_settings,
                        width=-1,
                        tag=sim_cdc_input,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Max. frequency".rjust(label_pad))
                    attach_tooltip(tooltips.simulation.max_freq)
                    dpg.add_input_float(
                        label="Hz",
                        default_value=state.config.default_simulation_settings.max_frequency,
                        min_value=1e-20,
                        max_value=1e20,
                        min_clamped=True,
                        max_clamped=True,
                        on_enter=True,
                        format="%.3E",
                        step=0.0,
                        width=-22,
                        callback=update_default_sim_settings,
                        tag=sim_max_freq_input,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Min. frequency".rjust(label_pad))
                    attach_tooltip(tooltips.simulation.min_freq)
                    dpg.add_input_float(
                        label="Hz",
                        default_value=state.config.default_simulation_settings.min_frequency,
                        min_value=1e-20,
                        max_value=1e20,
                        min_clamped=True,
                        max_clamped=True,
                        on_enter=True,
                        format="%.3E",
                        step=0.0,
                        width=-22,
                        callback=update_default_sim_settings,
                        tag=sim_min_freq_input,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Num. points per decade".rjust(label_pad))
                    attach_tooltip(tooltips.simulation.per_decade)
                    dpg.add_input_int(
                        default_value=state.config.default_simulation_settings.num_per_decade,
                        min_value=1,
                        min_clamped=True,
                        step=0,
                        on_enter=True,
                        callback=update_default_sim_settings,
                        tag=sim_per_decade_input,
                        width=-1,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("".rjust(label_pad))
                    dpg.add_button(
                        label="Restore defaults",
                        callback=restore_default_sim_settings,
                    )
            dpg.add_spacer(height=8)

        with dpg.collapsing_header(label="Export plot", default_open=True):
            export_units: int = dpg.generate_uuid()
            export_width: int = dpg.generate_uuid()
            export_height: int = dpg.generate_uuid()
            export_dpi: int = dpg.generate_uuid()
            export_preview: int = dpg.generate_uuid()
            export_title: int = dpg.generate_uuid()
            export_legend: int = dpg.generate_uuid()
            export_legend_location: int = dpg.generate_uuid()
            export_grid: int = dpg.generate_uuid()
            export_tight: int = dpg.generate_uuid()
            export_num_per_decade: int = dpg.generate_uuid()
            export_extension: int = dpg.generate_uuid()
            export_experimental_clear_registry: int = dpg.generate_uuid()
            export_experimental_disable_previews: int = dpg.generate_uuid()

            def restore_default_export_settings():
                state.config.export_units = 1
                state.config.export_width = 10.0
                state.config.export_height = 6.0
                state.config.export_dpi = 100
                state.config.export_preview = 4
                state.config.export_title = True
                state.config.export_legend = True
                state.config.export_legend_location = 0
                state.config.export_grid = False
                state.config.export_tight = False
                state.config.export_num_per_decade = 100
                state.config.export_extension = ".png"
                state.config.export_experimental_clear_registry = True
                state.config.export_experimental_disable_previews = False
                dpg.set_value(
                    export_units, list(UNITS_PER_INCH)[state.config.export_units]
                )
                dpg.set_value(export_width, state.config.export_width)
                dpg.set_value(export_height, state.config.export_height)
                dpg.set_value(export_dpi, state.config.export_dpi)
                dpg.set_value(
                    export_preview, list(PREVIEW_LIMITS)[state.config.export_preview]
                )
                dpg.set_value(export_title, state.config.export_title)
                dpg.set_value(export_legend, state.config.export_legend)
                dpg.set_value(
                    export_legend_location,
                    list(LEGEND_LOCATIONS)[state.config.export_legend_location],
                )
                dpg.set_value(export_grid, state.config.export_grid)
                dpg.set_value(export_tight, state.config.export_tight)
                dpg.set_value(export_num_per_decade, state.config.export_num_per_decade)
                dpg.set_value(export_extension, state.config.export_extension)
                dpg.set_value(
                    export_experimental_clear_registry,
                    state.config.export_experimental_clear_registry,
                )
                dpg.set_value(
                    export_experimental_disable_previews,
                    state.config.export_experimental_disable_previews,
                )
                state.plot_exporter.update_settings(state.config)

            def update_export_units(unit: str):
                state.config.export_units = list(UNITS_PER_INCH).index(unit)
                state.plot_exporter.update_settings(state.config)

            def update_export_width(width: float):
                state.config.export_width = width
                state.plot_exporter.update_settings(state.config)

            def update_export_height(height: float):
                state.config.export_height = height
                state.plot_exporter.update_settings(state.config)

            def update_export_dpi(dpi: float):
                state.config.export_dpi = dpi
                state.plot_exporter.update_settings(state.config)

            def update_export_preview(limit: str):
                state.config.export_preview = list(PREVIEW_LIMITS).index(limit)
                state.plot_exporter.update_settings(state.config)

            def update_export_title(flag: bool):
                state.config.export_title = flag
                state.plot_exporter.update_settings(state.config)

            def update_export_legend(flag: bool):
                state.config.export_legend = flag
                state.plot_exporter.update_settings(state.config)

            def update_export_legend_location(location: str):
                state.config.export_legend_location = LEGEND_LOCATIONS[location]
                state.plot_exporter.update_settings(state.config)

            def update_export_grid(flag: bool):
                state.config.export_grid = flag
                state.plot_exporter.update_settings(state.config)

            def update_export_tight(flag: bool):
                state.config.export_tight = flag
                state.plot_exporter.update_settings(state.config)

            def update_export_num_per_decade(num: int):
                state.config.export_num_per_decade = num
                state.plot_exporter.update_settings(state.config)

            def update_export_extension(ext: str):
                state.config.export_extension = EXTENSIONS.index(ext)
                state.plot_exporter.update_settings(state.config)

            def update_export_experimental_clear_registry(flag: bool):
                state.config.export_experimental_clear_registry = flag
                state.plot_exporter.update_settings(state.config)

            def update_export_experimental_disable_previews(flag: bool):
                state.config.export_experimental_disable_previews = flag
                state.plot_exporter.update_settings(state.config)

            with dpg.child_window(border=False, width=350, height=320):
                with dpg.group(horizontal=True):
                    dpg.add_text("Units".rjust(label_pad))
                    attach_tooltip(tooltips.plotting.export_units)
                    dpg.add_combo(
                        width=-1,
                        items=list(UNITS_PER_INCH.keys()),
                        default_value=list(UNITS_PER_INCH.keys())[
                            state.config.export_units
                        ],
                        callback=lambda s, a, u: update_export_units(a),
                        tag=export_units,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Width".rjust(label_pad))
                    attach_tooltip(tooltips.plotting.export_width)
                    dpg.add_input_float(
                        width=-1,
                        default_value=state.config.export_width,
                        min_value=0.01,
                        min_clamped=True,
                        step=0.0,
                        format="%.2f",
                        on_enter=True,
                        callback=lambda s, a, u: update_export_width(a),
                        tag=export_width,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Height".rjust(label_pad))
                    attach_tooltip(tooltips.plotting.export_height)
                    dpg.add_input_float(
                        width=-1,
                        default_value=state.config.export_height,
                        min_value=0.01,
                        min_clamped=True,
                        step=0.0,
                        format="%.2f",
                        on_enter=True,
                        callback=lambda s, a, u: update_export_height(a),
                        tag=export_height,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("DPI".rjust(label_pad))
                    attach_tooltip(tooltips.plotting.export_dpi)
                    dpg.add_input_int(
                        width=-1,
                        default_value=state.config.export_dpi,
                        min_value=1,
                        min_clamped=True,
                        step=0,
                        on_enter=True,
                        callback=lambda s, a, u: update_export_dpi(a),
                        tag=export_dpi,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Preview".rjust(label_pad))
                    attach_tooltip(tooltips.plotting.export_preview)
                    dpg.add_combo(
                        width=-1,
                        items=list(PREVIEW_LIMITS.keys()),
                        default_value=list(PREVIEW_LIMITS.keys())[
                            state.config.export_preview
                        ],
                        callback=lambda s, a, u: update_export_preview(a),
                        tag=export_preview,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Title".rjust(label_pad))
                    attach_tooltip(tooltips.plotting.export_title)
                    dpg.add_checkbox(
                        default_value=state.config.export_title,
                        callback=lambda s, a, u: update_export_title(a),
                        tag=export_title,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Legend".rjust(label_pad))
                    attach_tooltip(tooltips.plotting.export_legend)
                    dpg.add_checkbox(
                        default_value=state.config.export_legend,
                        callback=lambda s, a, u: update_export_legend(a),
                        tag=export_legend,
                    )
                    dpg.add_combo(
                        width=-1,
                        items=list(LEGEND_LOCATIONS.keys()),
                        default_value=list(LEGEND_LOCATIONS.keys())[0],
                        callback=lambda s, a, u: update_export_legend_location(a),
                        tag=export_legend_location,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Grid".rjust(label_pad))
                    attach_tooltip(tooltips.plotting.export_grid)
                    dpg.add_checkbox(
                        default_value=state.config.export_grid,
                        callback=lambda s, a, u: update_export_grid(a),
                        tag=export_grid,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Tight".rjust(label_pad))
                    attach_tooltip(tooltips.plotting.export_tight)
                    dpg.add_checkbox(
                        default_value=state.config.export_tight,
                        callback=lambda s, a, u: update_export_tight(a),
                        tag=export_tight,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Points per decade".rjust(label_pad))
                    attach_tooltip(tooltips.plotting.export_num_per_decade)
                    dpg.add_input_int(
                        width=-1,
                        default_value=state.config.export_num_per_decade,
                        min_value=1,
                        min_clamped=True,
                        step=0,
                        on_enter=True,
                        callback=lambda s, a, u: update_export_num_per_decade(a),
                        tag=export_num_per_decade,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Extension".rjust(label_pad))
                    attach_tooltip(tooltips.plotting.export_extension)
                    dpg.add_combo(
                        width=-1,
                        items=EXTENSIONS,
                        default_value=EXTENSIONS[state.config.export_extension],
                        callback=lambda s, a, u: update_export_extension(a),
                        tag=export_extension,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Clear texture registry".rjust(label_pad))
                    attach_tooltip(tooltips.plotting.export_experimental_clear_registry)
                    dpg.add_checkbox(
                        default_value=state.config.export_experimental_clear_registry,
                        callback=lambda s, a, u: update_export_experimental_clear_registry(
                            a
                        ),
                        tag=export_experimental_clear_registry,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("Disable plot previews".rjust(label_pad))
                    attach_tooltip(tooltips.plotting.export_experimental_disable_previews)
                    dpg.add_checkbox(
                        default_value=state.config.export_experimental_disable_previews,
                        callback=lambda s, a, u: update_export_experimental_disable_previews(
                            a
                        ),
                        tag=export_experimental_disable_previews,
                    )
                with dpg.group(horizontal=True):
                    dpg.add_text("".rjust(label_pad))
                    dpg.add_button(
                        label="Restore defaults",
                        callback=restore_default_export_settings,
                    )
    signals.emit(Signal.BLOCK_KEYBINDINGS, window=window)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from enum import IntEnum, auto
from typing import Dict, List


class Context(IntEnum):
    PROGRAM = auto()
    PROJECT = auto()
    OVERVIEW_TAB = auto()
    DATA_SETS_TAB = auto()
    KRAMERS_KRONIG_TAB = auto()
    FITTING_TAB = auto()
    SIMULATION_TAB = auto()
    PLOTTING_TAB = auto()


class Action(IntEnum):
    # Program-level
    NEW_PROJECT = auto()
    LOAD_PROJECT = auto()
    EXIT = auto()
    NEXT_PROGRAM_TAB = auto()
    PREVIOUS_PROGRAM_TAB = auto()
    SELECT_HOME_TAB = auto()
    SHOW_HELP_ABOUT = auto()
    SHOW_HELP_LICENSES = auto()
    SHOW_SETTINGS_APPEARANCE = auto()
    SHOW_SETTINGS_DEFAULTS = auto()
    SHOW_SETTINGS_KEYBINDINGS = auto()
    SHOW_COMMAND_PALETTE = auto()

    # Project-level
    SAVE_PROJECT = auto()
    SAVE_PROJECT_AS = auto()
    CLOSE_PROJECT = auto()
    UNDO = auto()
    REDO = auto()
    NEXT_PROJECT_TAB = auto()
    PREVIOUS_PROJECT_TAB = auto()
    SELECT_DATA_SETS_TAB = auto()
    SELECT_FITTING_TAB = auto()
    SELECT_KRAMERS_KRONIG_TAB = auto()
    SELECT_OVERVIEW_TAB = auto()
    SELECT_PLOTTING_TAB = auto()
    SELECT_SIMULATION_TAB = auto()

    # Project-level: multiple tabs
    PERFORM_ACTION = auto()
    # - Load data set
    # - Perform test
    # - Perform fit
    # - Perform simulation
    # - Create plot
    DELETE_RESULT = auto()
    # - Data set
    # - Test result
    # - Fit result
    # - Simulation result
    # - Plot
    NEXT_PRIMARY_RESULT = auto()
    PREVIOUS_PRIMARY_RESULT = auto()
    # - Data set
    # - Plot
    NEXT_SECONDARY_RESULT = auto()
    PREVIOUS_SECONDARY_RESULT = auto()
    # - Test result
    # - Fit result
    # - Simulation result
    # - Plot type
    APPLY_SETTINGS = auto()
    # - Kramers-Kronig tab
    # - Fitting tab
    # - Simulation tab
    APPLY_MASK = auto()
    # - Kramers-Kronig tab
    # - Fitting tab
    SHOW_ENLARGED_NYQUIST = auto()
    SHOW_ENLARGED_BODE = auto()
    SHOW_ENLARGED_RESIDUALS = auto()
    SHOW_CIRCUIT_EDITOR = auto()
    # - Fitting tab
    # - Simulation tab
    COPY_NYQUIST_DATA = auto()
    COPY_BODE_DATA = auto()
    COPY_RESIDUALS_DATA = auto()
    COPY_OUTPUT = auto()
    # - Fit output
    # - Simulation output

    # Project-level: data sets tab
    AVERAGE_DATA_SETS = auto()
    TOGGLE_DATA_POINTS = auto()
    COPY_DATA_SET_MASK = auto()
    SUBTRACT_IMPEDANCE = auto()

    # Project-level: plotting tab
    SELECT_ALL_PLOT_SERIES = auto()
    UNSELECT_ALL_PLOT_SERIES = auto()
    COPY_PLOT_APPEARANCE = auto()
    COPY_PLOT_DATA = auto()
    EXPAND_COLLAPSE_SIDEBAR = auto()
    EXPORT_PLOT = auto()


action_contexts: Dict[Action, List[Context]] = {
    Action.NEW_PROJECT: [Context.PROGRAM],
    Action.LOAD_PROJECT: [Context.PROGRAM],
    Action.EXIT: [Context.PROGRAM],
    Action.NEXT_PROGRAM_TAB: [Context.PROGRAM],
    Action.PREVIOUS_PROGRAM_TAB: [Context.PROGRAM],
    Action.SELECT_HOME_TAB: [Context.PROGRAM],
    Action.SHOW_HELP_ABOUT: [Context.PROGRAM],
    Action.SHOW_HELP_LICENSES: [Context.PROGRAM],
    Action.SHOW_SETTINGS_APPEARANCE: [Context.PROGRAM],
    Action.SHOW_SETTINGS_DEFAULTS: [Context.PROGRAM],
    Action.SHOW_SETTINGS_KEYBINDINGS: [Context.PROGRAM],
    Action.SHOW_COMMAND_PALETTE: [Context.PROGRAM],
    Action.SAVE_PROJECT: [Context.PROJECT],
    Action.SAVE_PROJECT_AS: [Context.PROJECT],
    Action.CLOSE_PROJECT: [Context.PROJECT],
    Action.UNDO: [Context.PROJECT],
    Action.REDO: [Context.PROJECT],
    Action.NEXT_PROJECT_TAB: [Context.PROJECT],
    Action.PREVIOUS_PROJECT_TAB: [Context.PROJECT],
    Action.SELECT_DATA_SETS_TAB: [Context.PROJECT],
    Action.SELECT_FITTING_TAB: [Context.PROJECT],
    Action.SELECT_KRAMERS_KRONIG_TAB: [Context.PROJECT],
    Action.SELECT_OVERVIEW_TAB: [Context.PROJECT],
    Action.SELECT_PLOTTING_TAB: [Context.PROJECT],
    Action.SELECT_SIMULATION_TAB: [Context.PROJECT],
    Action.PERFORM_ACTION: [
        Context.DATA_SETS_TAB,
        Context.KRAMERS_KRONIG_TAB,
        Context.FITTING_TAB,
        Context.SIMULATION_TAB,
        Context.PLOTTING_TAB,
    ],
    Action.DELETE_RESULT: [
        Context.DATA_SETS_TAB,
        Context.KRAMERS_KRONIG_TAB,
        Context.FITTING_TAB,
        Context.SIMULATION_TAB,
        Context.PLOTTING_TAB,
    ],
    Action.NEXT_PRIMARY_RESULT: [
        Context.DATA_SETS_TAB,
        Context.KRAMERS_KRONIG_TAB,
        Context.FITTING_TAB,
        Context.SIMULATION_TAB,
        Context.PLOTTING_TAB,
    ],
    Action.PREVIOUS_PRIMARY_RESULT: [
        Context.DATA_SETS_TAB,
        Context.KRAMERS_KRONIG_TAB,
        Context.FITTING_TAB,
        Context.SIMULATION_TAB,
        Context.PLOTTING_TAB,
    ],
    Action.NEXT_SECONDARY_RESULT: [
        Context.KRAMERS_KRONIG_TAB,
        Context.FITTING_TAB,
        Context.SIMULATION_TAB,
        Context.PLOTTING_TAB,
    ],
    Action.PREVIOUS_SECONDARY_RESULT: [
        Context.KRAMERS_KRONIG_TAB,
        Context.FITTING_TAB,
        Context.SIMULATION_TAB,
        Context.PLOTTING_TAB,
    ],
    Action.APPLY_SETTINGS: [
        Context.KRAMERS_KRONIG_TAB,
        Context.FITTING_TAB,
        Context.SIMULATION_TAB,
    ],
    Action.APPLY_MASK: [
        Context.KRAMERS_KRONIG_TAB,
        Context.FITTING_TAB,
    ],
    Action.SHOW_ENLARGED_NYQUIST: [
        Context.DATA_SETS_TAB,
        Context.KRAMERS_KRONIG_TAB,
        Context.FITTING_TAB,
        Context.SIMULATION_TAB,
    ],
    Action.SHOW_ENLARGED_BODE: [
        Context.DATA_SETS_TAB,
        Context.KRAMERS_KRONIG_TAB,
        Context.FITTING_TAB,
        Context.SIMULATION_TAB,
    ],
    Action.SHOW_ENLARGED_RESIDUALS: [
        Context.KRAMERS_KRONIG_TAB,
        Context.FITTING_TAB,
    ],
    Action.SHOW_CIRCUIT_EDITOR: [
        Context.FITTING_TAB,
        Context.SIMULATION_TAB,
    ],
    Action.COPY_NYQUIST_DATA: [
        Context.KRAMERS_KRONIG_TAB,
        Context.FITTING_TAB,
        Context.SIMULATION_TAB,
    ],
    Action.COPY_BODE_DATA: [
        Context.KRAMERS_KRONIG_TAB,
        Context.FITTING_TAB,
        Context.SIMULATION_TAB,
    ],
    Action.COPY_RESIDUALS_DATA: [
        Context.KRAMERS_KRONIG_TAB,
        Context.FITTING_TAB,
    ],
    Action.COPY_OUTPUT: [
        Context.FITTING_TAB,
        Context.SIMULATION_TAB,
    ],
    Action.AVERAGE_DATA_SETS: [Context.DATA_SETS_TAB],
    Action.TOGGLE_DATA_POINTS: [Context.DATA_SETS_TAB],
    Action.COPY_DATA_SET_MASK: [Context.DATA_SETS_TAB],
    Action.SUBTRACT_IMPEDANCE: [Context.DATA_SETS_TAB],
    Action.SELECT_ALL_PLOT_SERIES: [Context.PLOTTING_TAB],
    Action.UNSELECT_ALL_PLOT_SERIES: [Context.PLOTTING_TAB],
    Action.COPY_PLOT_APPEARANCE: [Context.PLOTTING_TAB],
    Action.COPY_PLOT_DATA: [Context.PLOTTING_TAB],
    Action.EXPAND_COLLAPSE_SIDEBAR: [Context.PLOTTING_TAB],
    Action.EXPORT_PLOT: [Context.PLOTTING_TAB],
}


action_to_string: Dict[Action, str] = {
    Action.NEW_PROJECT: "new-project",
    Action.LOAD_PROJECT: "load-project",
    Action.EXIT: "exit-program",
    Action.NEXT_PROGRAM_TAB: "next-program-tab",
    Action.PREVIOUS_PROGRAM_TAB: "previous-program-tab",
    Action.SELECT_HOME_TAB: "select-home-tab",
    Action.SHOW_HELP_ABOUT: "show-help-about",
    Action.SHOW_HELP_LICENSES: "show-help-licenses",
    Action.SHOW_SETTINGS_APPEARANCE: "show-settings-appearance",
    Action.SHOW_SETTINGS_DEFAULTS: "show-settings-defaults",
    Action.SHOW_SETTINGS_KEYBINDINGS: "show-settings-keybindings",
    Action.SHOW_COMMAND_PALETTE: "show-command-palette",
    Action.SAVE_PROJECT: "save-project",
    Action.SAVE_PROJECT_AS: "save-project-as",
    Action.CLOSE_PROJECT: "close-project",
    Action.UNDO: "undo",
    Action.REDO: "redo",
    Action.NEXT_PROJECT_TAB: "next-project-tab",
    Action.PREVIOUS_PROJECT_TAB: "previous-project-tab",
    Action.SELECT_DATA_SETS_TAB: "select-data-sets-tab",
    Action.SELECT_FITTING_TAB: "select-fitting-tab",
    Action.SELECT_KRAMERS_KRONIG_TAB: "select-kramers-kronig-tab",
    Action.SELECT_OVERVIEW_TAB: "select-overview-tab",
    Action.SELECT_PLOTTING_TAB: "select-plotting-tab",
    Action.SELECT_SIMULATION_TAB: "select-simulation-tab",
    Action.PERFORM_ACTION: "perform-action",
    Action.DELETE_RESULT: "delete-result",
    Action.NEXT_PRIMARY_RESULT: "next-primary-result",
    Action.PREVIOUS_PRIMARY_RESULT: "previous-primary-result",
    Action.NEXT_SECONDARY_RESULT: "next-secondary-result",
    Action.PREVIOUS_SECONDARY_RESULT: "previous-secondary-result",
    Action.APPLY_SETTINGS: "apply-settings",
    Action.APPLY_MASK: "apply-mask",
    Action.SHOW_ENLARGED_NYQUIST: "show-enlarged-nyquist",
    Action.SHOW_ENLARGED_BODE: "show-enlarged-bode",
    Action.SHOW_ENLARGED_RESIDUALS: "show-enlarged-residuals",
    Action.SHOW_CIRCUIT_EDITOR: "show-circuit-editor",
    Action.COPY_NYQUIST_DATA: "copy-nyquist-data",
    Action.COPY_BODE_DATA: "copy-bode-data",
    Action.COPY_RESIDUALS_DATA: "copy-residuals-data",
    Action.COPY_OUTPUT: "copy-output",
    Action.AVERAGE_DATA_SETS: "average-data-sets",
    Action.TOGGLE_DATA_POINTS: "toggle-data-points",
    Action.COPY_DATA_SET_MASK: "copy-data-set-mask",
    Action.SUBTRACT_IMPEDANCE: "subtract-impedance",
    Action.SELECT_ALL_PLOT_SERIES: "select-all-plot-series",
    Action.UNSELECT_ALL_PLOT_SERIES: "unselect-all-plot-series",
    Action.COPY_PLOT_APPEARANCE: "copy-plot-appearance",
    Action.COPY_PLOT_DATA: "copy-plot-data",
    Action.EXPAND_COLLAPSE_SIDEBAR: "expand-collapse-sidebar",
    Action.EXPORT_PLOT: "export-plot",
}
string_to_action: Dict[str, Action] = {v: k for k, v in action_to_string.items()}


action_descriptions: Dict[Action, str] = {
    Action.NEW_PROJECT: """
Create a new project.
""".strip(),
    Action.LOAD_PROJECT: """
Select project(s) to load.
""".strip(),
    Action.EXIT: """
Close the program.
""".strip(),
    Action.NEXT_PROGRAM_TAB: """
Go to the next project.
""".strip(),
    Action.PREVIOUS_PROGRAM_TAB: """
Go to the previous project.
""".strip(),
    Action.SELECT_HOME_TAB: """
Go to the 'Home' tab.
""".strip(),
    Action.SHOW_HELP_ABOUT: """
Show the 'About' window.
""".strip(),
    Action.SHOW_HELP_LICENSES: """
Show the 'Licenses' window.
""".strip(),
    Action.SHOW_SETTINGS_APPEARANCE: """
Show the 'Settings - appearance' window.
""".strip(),
    Action.SHOW_SETTINGS_DEFAULTS: """
Show the 'Settings - defaults' window.
""".strip(),
    Action.SHOW_SETTINGS_KEYBINDINGS: """
Show the 'Settings - keybindings' window.
""".strip(),
    Action.SHOW_COMMAND_PALETTE: """
Show the command palette, which can be used as an alternative to other keybindings for performing actions.
""".strip(),
    Action.SAVE_PROJECT: """
Save the current project.
""".strip(),
    Action.SAVE_PROJECT_AS: """
Save the current project as a new file.
""".strip(),
    Action.CLOSE_PROJECT: """
Close the current project.
""".strip(),
    Action.UNDO: """
Undo the latest action.
""".strip(),
    Action.REDO: """
Redo the latest action.
""".strip(),
    Action.NEXT_PROJECT_TAB: """
Go to the next project tab.
""".strip(),
    Action.PREVIOUS_PROJECT_TAB: """
Go to the previous project tab.
""".strip(),
    Action.SELECT_DATA_SETS_TAB: """
Go to the 'Data sets' tab.
""".strip(),
    Action.SELECT_FITTING_TAB: """
Go to the 'Fitting' tab.
""".strip(),
    Action.SELECT_KRAMERS_KRONIG_TAB: """
Go to the 'Kramers-Kronig' tab.
""".strip(),
    Action.SELECT_OVERVIEW_TAB: """
Go to the 'Overview' tab.
""".strip(),
    Action.SELECT_PLOTTING_TAB: """
Go to the 'Plotting' tab.
""".strip(),
    Action.SELECT_SIMULATION_TAB: """
Go to the 'Simulation' tab.
""".strip(),
    Action.PERFORM_ACTION: """
Perform the primary action of the current project tab:
- Data sets: select files to load.
- Kramers-Kronig: perform test.
- Fitting: perform fit.
- Simulation: perform simulation.
- Plotting: create a new plot.
""".strip(),
    Action.DELETE_RESULT: """
Delete the current result in the current project tab:
- Data sets: delete the current data set.
- Kramers-Kronig: delete the current test result.
- Fitting: delete the current fit result.
- Simulation: delete the current simulation result.
- Plotting: delete the current plot.
""".strip(),
    Action.NEXT_PRIMARY_RESULT: """
Select the next primary result of the current project tab:
- Data sets: data set.
- Kramers-Kronig: data set.
- Fitting: data set.
- Simulation: data set.
- Plotting: plot.
""".strip(),
    Action.PREVIOUS_PRIMARY_RESULT: """
Select the previous primary result of the current project tab:
- Data sets: data set.
- Kramers-Kronig: data set.
- Fitting: data set.
- Simulation: data set.
- Plotting: plot.
""".strip(),
    Action.NEXT_SECONDARY_RESULT: """
Select the next secondary result of the current project tab:
- Kramers-Kronig: test result.
- Fitting: fit result.
- Simulation: simulation result.
- Plotting: plot type.
""".strip(),
    Action.PREVIOUS_SECONDARY_RESULT: """
Select the previous secondary result of the current project tab:
- Kramers-Kronig: test result.
- Fitting: fit result.
- Simulation: simulation result.
- Plotting: plot type.
""".strip(),
    Action.APPLY_SETTINGS: """
Apply the settings used in the current secondary result of the current project tab:
- Kramers-Kronig: test result.
- Fitting: fit result.
- Simulation: simulation result.
""".strip(),
    Action.APPLY_MASK: """
Apply the mask used in the current secondary result of the current project tab:
- Kramers-Kronig: test result.
- Fitting: fit result.
""".strip(),
    Action.SHOW_ENLARGED_NYQUIST: """
Show an enlarged Nyquist plot.
""".strip(),
    Action.SHOW_ENLARGED_BODE: """
Show an enlarged Bode plot.
""".strip(),
    Action.SHOW_ENLARGED_RESIDUALS: """
Show an enlarged residuals plot.
""".strip(),
    Action.SHOW_CIRCUIT_EDITOR: """
Show the circuit editor window.
""".strip(),
    Action.COPY_NYQUIST_DATA: """
Copy the data from the Nyquist plot.
""".strip(),
    Action.COPY_BODE_DATA: """
Copy the data from the Bode plot.
""".strip(),
    Action.COPY_RESIDUALS_DATA: """
Copy the data from the residuals plot.
""".strip(),
    Action.COPY_OUTPUT: """
Copy the chosen output while in the 'Fitting' or the 'Simulation' tab.
""".strip(),
    Action.AVERAGE_DATA_SETS: """
Select data sets to average.
""".strip(),
    Action.TOGGLE_DATA_POINTS: """
Select data points to toggle.
""".strip(),
    Action.COPY_DATA_SET_MASK: """
Select which data set's mask to copy.
""".strip(),
    Action.SUBTRACT_IMPEDANCE: """
Select the impedance to subtract from the current data set.
""".strip(),
    Action.SELECT_ALL_PLOT_SERIES: """
Select all data sets and test/fit/simulation results.
""".strip(),
    Action.UNSELECT_ALL_PLOT_SERIES: """
Unselect all data sets and test/fit/simulation results.
""".strip(),
    Action.COPY_PLOT_APPEARANCE: """
Select which plot's appearance settings to copy.
""".strip(),
    Action.COPY_PLOT_DATA: """
Copy the data from the current plot while in the 'Plotting' tab.
""".strip(),
    Action.EXPAND_COLLAPSE_SIDEBAR: """
Collapse/expand the sidebar while in the 'Plotting' tab.
""".strip(),
    Action.EXPORT_PLOT: """
Export the current plot using matplotlib.
""".strip(),
}


class Method(IntEnum):
    """
    Iterative methods used during complex non-linear least-squares fitting:

    - AUTO: try each method
    - AMPGO
    - BASINHOPPING
    - BFGS
    - BRUTE
    - CG
    - COBYLA
    - DIFFERENTIAL_EVOLUTION
    - DOGLEG
    - DUAL_ANNEALING
    - EMCEE
    - LBFGSB
    - LEASTSQ
    - LEAST_SQUARES
    - NELDER
    - NEWTON
    - POWELL
    - SHGO
    - SLSQP
    - TNC
    - TRUST_CONSTR
    - TRUST_EXACT
    - TRUST_KRYLOV
    - TRUST_NCG
    """

    AUTO = 1
    LEASTSQ = 2
    LEAST_SQUARES = 3
    DIFFERENTIAL_EVOLUTION = 4
    BRUTE = 5
    BASINHOPPING = 6
    AMPGO = 7
    NELDER = 8
    LBFGSB = 9
    POWELL = 10
    CG = 11
    NEWTON = 12
    COBYLA = 13
    BFGS = 14
    TNC = 15
    TRUST_NCG = 16
    TRUST_EXACT = 17
    TRUST_KRYLOV = 18
    TRUST_CONSTR = 19
    DOGLEG = 20
    SLSQP = 21
    EMCEE = 22
    SHGO = 23
    DUAL_ANNEALING = 24


label_to_method: Dict[str, Method] = {
    "Levenberg-Marquardt": Method.LEASTSQ,
    "Least-squares (trust-region reflective)": Method.LEAST_SQUARES,
    "Powell": Method.POWELL,  # Had some problem fitting
    "BFGS": Method.BFGS,  # Had some problem fitting
    "Nelder-Mead": Method.NELDER,  # Had some problem fitting
    "Conjugate Gradient": Method.CG,  # Had some problem fitting
    "Truncated Newton": Method.TNC,  # Had some problem fitting
    "L-BFGS-B": Method.LBFGSB,  # NaN errors
    # "COBYLA": Method.COBYLA,  # Had some problem fitting
    "Sequential Linear Squares Programming": Method.SLSQP,  # Had some problem fitting
    # "Basin hopping": Method.BASINHOPPING,  # NaN errors
    # "Differential Evolution": Method.DIFFERENTIAL_EVOLUTION,  # Requires finite bounds for all mutable parameters
    # "Brute force method": Method.BRUTE,  # Requires that brute_step is defined for mutable parameters
    # "Adaptive Memory Programming for Global Optimization": Method.AMPGO,  # NaN errors
    # "Newton": Method.NEWTON,  # Requires Jacobian
    # "Newton CG": Method.TRUST_NCG,  # Requires Jacobian for trust region
    # "Exact trust-region": Method.TRUST_EXACT,  # Requires Jacobian for trust region
    # "Newton GLTR trust-region": Method.TRUST_KRYLOV,  # Requires Jacobian for trust region
    # "Constrained trust-region": Method.TRUST_CONSTR,  # NaN errors
    # "Dog-leg trust-region": Method.DOGLEG,  # Requires Jacobian
    # "Simplicial Homology Global Optimization": Method.SHGO,  # NaN errors
    # "Dual Annealing": Method.DUAL_ANNEALING,  # Requires finite bounds for all mutable parameters
    # "Maximum likelyhood via Monte-Carlo Markov chain": Method.EMCEE,  # Requires the emcee package (version 3)
}
method_to_label: Dict[Method, str] = {v: k for k, v in label_to_method.items()}
method_to_label[Method.AUTO] = "Auto"
method_to_value: Dict[Method, str] = {
    Method.AUTO: "auto",
    Method.LEASTSQ: "leastsq",
    Method.LEAST_SQUARES: "least_squares",
    Method.DIFFERENTIAL_EVOLUTION: "differential_evolution",
    Method.BRUTE: "brute",
    Method.BASINHOPPING: "basinhopping",
    Method.AMPGO: "ampgo",
    Method.NELDER: "nelder",
    Method.LBFGSB: "lbfgsb",
    Method.POWELL: "powell",
    Method.CG: "cg",
    Method.NEWTON: "newton",
    Method.COBYLA: "cobyla",
    Method.BFGS: "bfgs",
    Method.TNC: "tnc",
    Method.TRUST_NCG: "trust-ncg",
    Method.TRUST_EXACT: "trust-exact",
    Method.TRUST_KRYLOV: "trust-krylov",
    Method.TRUST_CONSTR: "trust-constr",
    Method.DOGLEG: "dogleg",
    Method.SLSQP: "slsqp",
    Method.EMCEE: "emcee",
    Method.SHGO: "shgo",
    Method.DUAL_ANNEALING: "dual_annealing",
}
value_to_method: Dict[str, Method] = {v: k for k, v in method_to_value.items()}


class Mode(IntEnum):
    """
    Types of modes that determine how the number of Voigt elements (capacitor connected in parallel with resistor) is chosen:

    - AUTO: follow procedure described by Schönleber, Klotz, and Ivers-Tiffée (2014)
    - EXPLORATORY: same procedure as AUTO but present intermediate results to user and apply additional weighting to the initial suggestion
    - MANUAL: manually choose the number
    """

    AUTO = 1
    EXPLORATORY = 2
    MANUAL = 3


label_to_mode: Dict[str, Mode] = {
    "Auto": Mode.AUTO,
    "Exploratory": Mode.EXPLORATORY,
    "Manual": Mode.MANUAL,
}
mode_to_label: Dict[Mode, str] = {v: k for k, v in label_to_mode.items()}


class Output(IntEnum):
    CDC_BASIC = auto()
    CDC_EXTENDED = auto()
    CSV_DATA_TABLE = auto()
    CSV_PARAMETERS_TABLE = auto()
    JSON_PARAMETERS_TABLE = auto()
    LATEX_DIAGRAM = auto()
    LATEX_EXPR = auto()
    LATEX_PARAMETERS_TABLE = auto()
    MARKDOWN_PARAMETERS_TABLE = auto()
    SYMPY_EXPR = auto()
    SYMPY_EXPR_VALUES = auto()


label_to_output: Dict[str, Output] = {
    "CDC - basic": Output.CDC_BASIC,
    "CDC - extended": Output.CDC_EXTENDED,
    "CSV - impedance table": Output.CSV_DATA_TABLE,
    "CSV - parameters table": Output.CSV_PARAMETERS_TABLE,
    "JSON - parameters table": Output.JSON_PARAMETERS_TABLE,
    "LaTeX - circuit diagram": Output.LATEX_DIAGRAM,
    "LaTeX - expression": Output.LATEX_EXPR,
    "LaTeX - parameters table": Output.LATEX_PARAMETERS_TABLE,
    "Markdown - parameters table": Output.MARKDOWN_PARAMETERS_TABLE,
    "SymPy - expression": Output.SYMPY_EXPR,
    "SymPy - expression and values": Output.SYMPY_EXPR_VALUES,
}
output_to_label: Dict[Output, str] = {v: k for k, v in label_to_output.items()}


class PlotType(IntEnum):
    """
    Types of plots:

    - NYQUIST: -Zim vs Zre
    - BODE_MAGNITUDE: |Z| vs log f
    - BODE_PHASE: phi vs log f
    """

    NYQUIST = 1
    BODE_MAGNITUDE = 2
    BODE_PHASE = 3


label_to_plot_type: Dict[str, PlotType] = {
    "Nyquist": PlotType.NYQUIST,
    "Bode - magnitude": PlotType.BODE_MAGNITUDE,
    "Bode - phase": PlotType.BODE_PHASE,
}
plot_type_to_label: Dict[PlotType, str] = {v: k for k, v in label_to_plot_type.items()}


class Test(IntEnum):
    """
    Types of tests:

    - CNLS: complex non-linear least-squares fit of circuit (fig. 1, Boukamp, 1995) with a distribution of fixed time constants
    - COMPLEX: eqs. 11 and 12, Boukamp, 1995
    - IMAGINARY: eqs. 4, 6, and 7, Boukamp, 1995
    - REAL: eqs. 5, 8, 9, and 10, Boukamp, 1995
    """

    CNLS = 1
    COMPLEX = 2
    IMAGINARY = 3
    REAL = 4


label_to_test: Dict[str, Test] = {
    "CNLS": Test.CNLS,
    "Complex": Test.COMPLEX,
    "Imaginary": Test.IMAGINARY,
    "Real": Test.REAL,
}
test_to_label: Dict[Test, str] = {v: k for k, v in label_to_test.items()}
test_to_value: Dict[Test, str] = {
    Test.CNLS: "cnls",
    Test.COMPLEX: "complex",
    Test.IMAGINARY: "imaginary",
    Test.REAL: "real",
}


class Weight(IntEnum):
    """
    Types of weights:

    - AUTO: try each weight
    - BOUKAMP: 1 / (Zre^2 + Zim^2) (eq. 13, Boukamp, 1995)
    - MODULUS: 1 / |Z|
    - PROPORTIONAL: 1 / Zre^2, 1 / Zim^2
    - UNITY: 1
    """

    AUTO = 1
    UNITY = 2
    PROPORTIONAL = 3
    MODULUS = 4
    BOUKAMP = 5


label_to_weight: Dict[str, Weight] = {
    "Modulus": Weight.MODULUS,
    "Proportional": Weight.PROPORTIONAL,
    "Unity": Weight.UNITY,
    "Boukamp": Weight.BOUKAMP,
}
weight_to_label: Dict[Weight, str] = {v: k for k, v in label_to_weight.items()}
weight_to_label[Weight.AUTO] = "Auto"
weight_to_value: Dict[Weight, str] = {
    Weight.AUTO: "auto",
    Weight.UNITY: "unity",
    Weight.PROPORTIONAL: "proportional",
    Weight.MODULUS: "modulus",
    Weight.BOUKAMP: "boukamp",
}
value_to_weight: Dict[str, Weight] = {v: k for k, v in weight_to_value.items()}



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

import dearpygui.dearpygui as dpg
from deareis.signals import Signal
import deareis.signals as signals
from deareis.utility import calculate_window_position_dimensions
from .base import Plot
from .bode import (
    Bode,
    BodeMagnitude,
    BodePhase,
)
from .image import Image
from .mu_xps import MuXps
from .nyquist import Nyquist
from .residuals import Residuals


def show_modal_plot_window(original: Plot, adjust_limits: bool = True):
    assert hasattr(original, "duplicate"), type(original)
    labels: dict = {
        Bode: "Bode",
        BodeMagnitude: "Bode - magnitude",
        BodePhase: "Bode - phase",
        Nyquist: "Nyquist",
        Residuals: "Residuals",
    }
    x: int
    y: int
    w: int
    h: int
    x, y, w, h = calculate_window_position_dimensions()
    window: int = dpg.generate_uuid()
    key_handler = dpg.generate_uuid()

    def close():
        dpg.delete_item(window)
        dpg.delete_item(key_handler)
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)

    with dpg.window(
        label=labels.get(type(original), "Unknown plot type"),
        modal=True,
        pos=(
            x,
            y,
        ),
        width=w,
        height=h,
        on_close=close,
        tag=window,
    ):
        copy: Plot = type(original).duplicate(original)
        if adjust_limits:
            copy.queue_limits_adjustment()
        else:
            copy.copy_limits(original)
    with dpg.handler_registry(tag=key_handler):
        dpg.add_key_release_handler(key=dpg.mvKey_Escape, callback=close)
    signals.emit(Signal.BLOCK_KEYBINDINGS, window=window)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import Callable, Dict, List, Optional, Tuple, Union
import dearpygui.dearpygui as dpg
from deareis.signals import Signal
import deareis.signals as signals
from deareis.data import DataSet, PlotSettings, TestResult, FitResult, SimulationResult
from deareis.utility import calculate_window_position_dimensions
from deareis.tooltips import attach_tooltip
import deareis.tooltips as tooltips
from deareis.data import Project
import deareis.themes as themes
from deareis.keybindings import (
    is_alt_down,
    is_control_down,
)


class SeriesBefore:
    def __init__(
        self,
        series: Union[DataSet, TestResult, FitResult, SimulationResult],
        settings: PlotSettings,
        marker_lookup: Dict[int, str],
        toggle_callback: Callable,
    ):
        assert type(series) in [
            DataSet,
            TestResult,
            FitResult,
            SimulationResult,
        ], series
        assert type(settings) is PlotSettings, settings
        assert type(marker_lookup) is dict, marker_lookup
        with dpg.table_row(user_data=series.uuid):
            dpg.add_checkbox(
                default_value=True,
                callback=lambda s, a, u: toggle_callback(a, u),
                user_data=series.uuid,
            )
            attach_tooltip(tooltips.plotting.copy_appearance_toggle)
            label = settings.get_series_label(series.uuid) or series.get_label()
            dpg.add_text(label)
            attach_tooltip(
                label
                + ("" if label == series.get_label() else f"\n{series.get_label()}")
            )
            dpg.add_color_edit(
                default_value=settings.get_series_color(series.uuid),
                enabled=False,
                no_picker=True,
                alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                no_inputs=True,
            )
            dpg.add_text(
                marker_lookup.get(settings.get_series_marker(series.uuid), "None")
            )
            dpg.add_checkbox(
                default_value=settings.get_series_line(series.uuid),
                enabled=False,
            )


class SeriesAfter:
    def __init__(
        self,
        series,
        source,
        destination,
        copy_labels,
        copy_colors,
        copy_markers,
        copy_lines,
        marker_lookup,
        parent,
    ):
        with dpg.table_row(parent=parent, user_data=series.uuid):
            label: str = (
                source.get_series_label(series.uuid)
                if copy_labels
                else destination.get_series_label(series.uuid)
            )
            dpg.add_text(
                label or series.get_label(),
                user_data=label,
            )
            attach_tooltip(
                label
                or series.get_label()
                + (
                    ""
                    if (label or series.get_label()) == series.get_label()
                    else f"\n{series.get_label()}"
                )
            )
            color: List[float] = (
                source.get_series_color(series.uuid)
                if copy_colors
                else destination.get_series_color(series.uuid)
            )
            dpg.add_color_edit(
                default_value=color,
                enabled=False,
                no_picker=True,
                alpha_preview=dpg.mvColorEdit_AlphaPreviewHalf,
                no_inputs=True,
                user_data=color,
            )
            marker: int = (
                source.get_series_marker(series.uuid)
                if copy_markers
                else destination.get_series_marker(series.uuid)
            )
            dpg.add_text(
                marker_lookup.get(marker, "None"),
                user_data=marker,
            )
            show_line: bool = (
                source.get_series_line(series.uuid)
                if copy_lines
                else destination.get_series_line(series.uuid)
            )
            dpg.add_checkbox(
                default_value=show_line,
                enabled=False,
                user_data=show_line,
            )


class CopyPlotAppearance:
    def __init__(self, settings: PlotSettings, project: Project):
        assert type(settings) is PlotSettings, settings
        assert type(project) is Project, project
        self.settings: PlotSettings = settings
        self.marker_lookup: Dict[int, str] = {
            v: k for k, v in themes.PLOT_MARKERS.items()
        }
        self.plot_lookup: Dict[str, PlotSettings] = {
            _.get_label(): _ for _ in project.get_plots() if _ != settings
        }
        self.labels: List[str] = list(sorted(self.plot_lookup.keys()))
        self.series_checkboxes: Dict[str, bool] = {
            _: True for _ in settings.series_order
        }
        self.data_sets: List[DataSet] = project.get_data_sets()
        self.tests: Dict[str, List[TestResult]] = project.get_all_tests()
        self.fits: Dict[str, List[FitResult]] = project.get_all_fits()
        self.simulations: List[SimulationResult] = project.get_simulations()
        x: int
        y: int
        w: int
        h: int
        x, y, w, h = calculate_window_position_dimensions(720, 540)
        self.window: int = dpg.generate_uuid()
        with dpg.window(
            label="Copy appearance settings",
            modal=True,
            no_resize=True,
            pos=(
                x,
                y,
            ),
            width=w,
            height=h,
            on_close=self.close,
            tag=self.window,
        ):
            with dpg.group(horizontal=True):
                dpg.add_text("Source")
                self.source_combo: int = dpg.generate_uuid()
                dpg.add_combo(
                    items=self.labels,
                    default_value=self.labels[0],
                    width=-1,
                    callback=self.change_source,
                    tag=self.source_combo,
                )
            with dpg.group(horizontal=True):
                with dpg.group():
                    dpg.add_text("Before")
                    with dpg.child_window(width=348, height=-24):
                        with dpg.table(
                            borders_outerV=True,
                            borders_outerH=True,
                            borders_innerV=True,
                            borders_innerH=True,
                            scrollY=True,
                            freeze_rows=1,
                            width=-1,
                            height=18 + 23 * len(self.settings.series_order),
                        ):
                            dpg.add_table_column(label="", width_fixed=True)
                            dpg.add_table_column(label="Label")
                            dpg.add_table_column(label="Color", width_fixed=True)
                            dpg.add_table_column(label="Marker", width_fixed=True)
                            dpg.add_table_column(label="Line", width_fixed=True)
                            uuid: str
                            for uuid in settings.series_order:
                                series: Optional[
                                    Union[
                                        DataSet, TestResult, FitResult, SimulationResult
                                    ]
                                ]
                                series = settings.find_series(
                                    uuid,
                                    self.data_sets,
                                    self.tests,
                                    self.fits,
                                    self.simulations,
                                )
                                assert series is not None
                                SeriesBefore(
                                    series,
                                    settings,
                                    self.marker_lookup,
                                    self.toggle_series,
                                )
                with dpg.group():
                    dpg.add_text("After")
                    with dpg.child_window(width=348, height=-24):
                        self.preview_table: int = dpg.generate_uuid()
                        with dpg.table(
                            borders_outerV=True,
                            borders_outerH=True,
                            borders_innerV=True,
                            borders_innerH=True,
                            scrollY=True,
                            freeze_rows=1,
                            width=-1,
                            height=18 + 23 * len(self.settings.series_order),
                            tag=self.preview_table,
                        ):
                            dpg.add_table_column(label="Label")
                            dpg.add_table_column(label="Color", width_fixed=True)
                            dpg.add_table_column(label="Marker", width_fixed=True)
                            dpg.add_table_column(label="Line", width_fixed=True)
            with dpg.group(horizontal=True):
                dpg.add_button(
                    label="Accept",
                    callback=self.accept_settings,
                )
                dpg.add_spacer(width=354)
                self.labels_checkbox: int = dpg.generate_uuid()
                dpg.add_checkbox(
                    label="Labels",
                    default_value=True,
                    tag=self.labels_checkbox,
                    callback=lambda s, a, u: self.change_source(),
                )
                attach_tooltip(tooltips.plotting.copy_appearance_labels)
                self.colors_checkbox: int = dpg.generate_uuid()
                dpg.add_checkbox(
                    label="Colors",
                    default_value=True,
                    tag=self.colors_checkbox,
                    callback=lambda s, a, u: self.change_source(),
                )
                attach_tooltip(tooltips.plotting.copy_appearance_colors)
                self.markers_checkbox: int = dpg.generate_uuid()
                dpg.add_checkbox(
                    label="Markers",
                    default_value=True,
                    tag=self.markers_checkbox,
                    callback=lambda s, a, u: self.change_source(),
                )
                attach_tooltip(tooltips.plotting.copy_appearance_markers)
                self.lines_checkbox: int = dpg.generate_uuid()
                dpg.add_checkbox(
                    label="Lines",
                    default_value=True,
                    tag=self.lines_checkbox,
                    callback=lambda s, a, u: self.change_source(),
                )
                attach_tooltip(tooltips.plotting.copy_appearance_lines)
        self.key_handler: int = dpg.generate_uuid()
        with dpg.handler_registry(tag=self.key_handler):
            dpg.add_key_release_handler(
                key=dpg.mvKey_Escape,
                callback=self.close,
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Return,
                callback=lambda: self.accept_settings(keybinding=True),
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Prior,
                callback=lambda: self.cycle_source(-1),
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Next,
                callback=lambda: self.cycle_source(1),
            )
        signals.emit(Signal.BLOCK_KEYBINDINGS, window=self.window)
        self.change_source()

    def close(self):
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)
        dpg.delete_item(self.window)
        dpg.delete_item(self.key_handler)

    def accept_settings(self, keybinding: bool = False):
        if keybinding is True and not (
            is_control_down()
            if dpg.get_platform() == dpg.mvPlatform_Windows
            else is_alt_down()
        ):
            return
        dpg.hide_item(self.window)
        changes: Dict[str, Tuple[str, List[float], int, bool]] = {}
        row: int
        for row in dpg.get_item_children(self.preview_table, slot=1):
            uuid: str = dpg.get_item_user_data(row)
            label: str = ""
            color: List[float] = [255.0, 255.0, 255.0, 255.0]
            marker: int = -1
            line: bool = False
            i: int
            item: int
            for i, item in enumerate(dpg.get_item_children(row, slot=1)):
                item_type: str = dpg.get_item_type(item)
                if i == 0:
                    assert "::mvText" in item_type, item_type
                    label = dpg.get_item_user_data(item)
                elif i == 1:
                    assert "::mvTooltip" in item_type, item_type
                elif i == 2:
                    assert "::mvColorEdit" in item_type, item_type
                    color = dpg.get_item_user_data(item)
                elif i == 3:
                    assert "::mvText" in item_type, item_type
                    marker = dpg.get_item_user_data(item)
                elif i == 4:
                    assert "::mvCheckbox" in item_type, item_type
                    line = dpg.get_item_user_data(item)
            changes[uuid] = (
                label,
                color,
                marker,
                line,
            )
        signals.emit(
            Signal.COPY_PLOT_APPEARANCE_SETTINGS,
            changes=changes,
            settings=self.settings,
        )
        self.close()

    def cycle_source(self, step: int):
        assert type(step) is int, step
        index: int = self.labels.index(dpg.get_value(self.source_combo)) + step
        dpg.set_value(self.source_combo, self.labels[index % len(self.labels)])
        self.change_source()

    def change_source(self):
        label: str = dpg.get_value(self.source_combo)
        source: PlotSettings = self.plot_lookup[label]
        dpg.delete_item(self.preview_table, children_only=True, slot=1)
        copy_labels: bool = dpg.get_value(self.labels_checkbox)
        copy_colors: bool = dpg.get_value(self.colors_checkbox)
        copy_markers: bool = dpg.get_value(self.markers_checkbox)
        copy_lines: bool = dpg.get_value(self.lines_checkbox)
        for uuid in self.settings.series_order:
            SeriesAfter(
                self.settings.find_series(
                    uuid, self.data_sets, self.tests, self.fits, self.simulations
                ),
                source
                if uuid in source.themes and self.series_checkboxes[uuid]
                else self.settings,
                self.settings,
                copy_labels,
                copy_colors,
                copy_markers,
                copy_lines,
                self.marker_lookup,
                self.preview_table,
            )

    def toggle_series(self, state: bool, uuid: str):
        self.series_checkboxes[uuid] = state
        self.change_source()



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import Callable, Dict, List, Optional
from numpy import angle, array, ndarray
import dearpygui.dearpygui as dpg
from deareis.signals import Signal
import deareis.signals as signals
from deareis.gui.plots import Bode, Nyquist
from deareis.utility import align_numbers, format_number
from deareis.tooltips import attach_tooltip, update_tooltip
import deareis.tooltips as tooltips
import deareis.themes as themes
from deareis.data import DataSet


class DataTable:
    def __init__(self):
        self._table: int = dpg.generate_uuid()
        with dpg.table(
            borders_outerV=True,
            borders_outerH=True,
            borders_innerV=True,
            borders_innerH=True,
            scrollY=True,
            freeze_rows=1,
            tag=self._table,
        ):
            dpg.add_table_column(
                label="Mask",
                width_fixed=True,
            )
            attach_tooltip(tooltips.data_sets.mask)
            dpg.add_table_column(
                label="Index",
                width_fixed=True,
            )
            dpg.add_table_column(
                label="f (Hz)",
            )
            attach_tooltip(tooltips.data_sets.frequency)
            dpg.add_table_column(
                label="Z' (ohm)",
            )
            attach_tooltip(tooltips.data_sets.real)
            dpg.add_table_column(
                label='-Z" (ohm)',
            )
            attach_tooltip(tooltips.data_sets.imaginary)
            dpg.add_table_column(
                label="|Z| (ohm)",
            )
            attach_tooltip(tooltips.data_sets.magnitude)
            dpg.add_table_column(
                label="-phi (°)",
            )
            attach_tooltip(tooltips.data_sets.phase)

    def clear(self):
        dpg.delete_item(self._table, children_only=True, slot=1)

    def populate(self, data: DataSet):
        assert type(data) is DataSet, data
        num_rows: int = data.get_num_points(masked=None)
        if num_rows == 0:
            self.clear()
            return
        rows: List[int] = dpg.get_item_children(self._table, slot=1)
        if len(rows) > num_rows:
            while len(rows) > num_rows:
                dpg.delete_item(rows.pop())
            return
        i: int
        for i in range(0, num_rows - len(rows)):
            with dpg.table_row(parent=self._table):
                dpg.add_checkbox(
                    default_value=False,
                    callback=lambda s, a, u: signals.emit(
                        Signal.TOGGLE_DATA_POINT,
                        state=a,
                        index=u[0],
                        data=u[1],
                    ),
                    user_data=(
                        i,
                        data,
                    ),
                )
                dpg.add_text("")
                dpg.add_text("")
                attach_tooltip("")
                dpg.add_text("")
                attach_tooltip("")
                dpg.add_text("")
                attach_tooltip("")
                dpg.add_text("")
                attach_tooltip("")
                dpg.add_text("")
                attach_tooltip("")

    def update(self, data: DataSet):
        assert type(data) is DataSet, data
        rows: List[int] = dpg.get_item_children(self._table, slot=1)
        assert len(rows) == data.get_num_points(masked=None)
        mask: Dict[int, bool] = data.get_mask()
        indices: List[str] = list(
            map(lambda _: str(_ + 1), range(0, data.get_num_points(masked=None)))
        )
        frequencies: ndarray = data.get_frequency(masked=None)
        freqs: List[str] = list(
            map(
                lambda _: format_number(_, significants=4),
                frequencies,
            )
        )
        Z: ndarray = data.get_impedance(masked=None)
        reals: List[str] = list(map(lambda _: format_number(_.real, significants=4), Z))
        imags: List[str] = list(
            map(lambda _: format_number(-_.imag, significants=4), Z)
        )
        mags: List[str] = list(map(lambda _: format_number(abs(_), significants=4), Z))
        phis: List[str] = list(
            map(
                lambda _: format_number(
                    -angle(_, deg=True),  # type: ignore
                    significants=4,
                    exponent=False,
                ),
                Z,
            )
        )
        indices = align_numbers(indices)
        freqs = align_numbers(freqs)
        reals = align_numbers(reals)
        imags = align_numbers(imags)
        mags = align_numbers(mags)
        phis = align_numbers(phis)
        fmt: str = "{:.6E}"
        i: int
        row: int
        for i, row in enumerate(rows):
            cells: List[int] = dpg.get_item_children(row, slot=1)
            # Mask
            dpg.configure_item(
                cells[0],
                default_value=mask.get(i, False),
                user_data=(
                    i,
                    data,
                ),
            )
            # Index
            dpg.configure_item(
                cells[1],
                default_value=indices[i],
            )
            # Frequency
            dpg.configure_item(
                cells[2],
                default_value=freqs[i],
            )
            update_tooltip(dpg.get_item_user_data(cells[3]), fmt.format(frequencies[i]))
            # Z, real
            dpg.configure_item(
                cells[4],
                default_value=reals[i],
            )
            update_tooltip(dpg.get_item_user_data(cells[5]), fmt.format(Z[i].real))
            # -Z, imaginary
            dpg.configure_item(
                cells[6],
                default_value=imags[i],
            )
            update_tooltip(dpg.get_item_user_data(cells[7]), fmt.format(-Z[i].imag))
            # Z, magnitude
            dpg.configure_item(
                cells[8],
                default_value=mags[i],
            )
            update_tooltip(dpg.get_item_user_data(cells[9]), fmt.format(abs(Z[i])))
            # Phase shift
            dpg.configure_item(
                cells[10],
                default_value=phis[i],
            )
            update_tooltip(
                dpg.get_item_user_data(cells[11]), fmt.format(-angle(Z[i], deg=True))
            )


class DataSetsTab:
    def __init__(self):
        self.queued_update: Optional[Callable] = None
        self.tab: int = dpg.generate_uuid()
        with dpg.tab(label="Data sets", tag=self.tab):
            with dpg.child_window(border=False, width=-1, height=-1):
                with dpg.group(horizontal=True):
                    self.table_window: int = dpg.generate_uuid()
                    self.table_width: int = 600
                    with dpg.child_window(
                        border=False,
                        width=self.table_width,
                        tag=self.table_window,
                        show=True,
                    ):
                        label_pad: int = 8
                        with dpg.child_window(border=True, height=82, width=-2):
                            with dpg.group(horizontal=True):
                                with dpg.child_window(border=False, width=-72):
                                    # TODO: Split into combo class?
                                    self.data_sets_combo: int = dpg.generate_uuid()
                                    with dpg.group(horizontal=True):
                                        self.visibility_item: int = dpg.generate_uuid()
                                        dpg.add_text(
                                            "Data set".rjust(label_pad),
                                            tag=self.visibility_item,
                                        )
                                        dpg.add_combo(
                                            callback=lambda s, a, u: signals.emit(
                                                Signal.SELECT_DATA_SET,
                                                data=u.get(a),
                                            ),
                                            user_data={},
                                            width=-1,
                                            tag=self.data_sets_combo,
                                        )
                                    self.label_input: int = dpg.generate_uuid()
                                    with dpg.group(horizontal=True):
                                        dpg.add_text("Label".rjust(label_pad))
                                        dpg.add_input_text(
                                            on_enter=True,
                                            callback=lambda s, a, u: signals.emit(
                                                Signal.RENAME_DATA_SET,
                                                label=a,
                                                data=u,
                                            ),
                                            width=-1,
                                            tag=self.label_input,
                                        )
                                    self.path_input: int = dpg.generate_uuid()
                                    with dpg.group(horizontal=True):
                                        dpg.add_text("Path".rjust(label_pad))
                                        dpg.add_input_text(
                                            on_enter=True,
                                            callback=lambda s, a, u: signals.emit(
                                                Signal.MODIFY_DATA_SET_PATH,
                                                path=a,
                                                data=u,
                                            ),
                                            width=-1,
                                            tag=self.path_input,
                                        )
                                with dpg.child_window(border=False, width=-1):
                                    self.load_button: int = dpg.generate_uuid()
                                    dpg.add_button(
                                        label="Load",
                                        callback=lambda s, a, u: signals.emit(
                                            Signal.SELECT_DATA_SET_FILES,
                                        ),
                                        width=-1,
                                        tag=self.load_button,
                                    )
                                    attach_tooltip(tooltips.data_sets.load)
                                    self.delete_button: int = dpg.generate_uuid()
                                    dpg.add_button(
                                        label="Delete",
                                        callback=lambda s, a, u: signals.emit(
                                            Signal.DELETE_DATA_SET,
                                            data=u,
                                        ),
                                        width=-1,
                                        tag=self.delete_button,
                                    )
                                    attach_tooltip(tooltips.data_sets.delete)
                                    self.average_button: int = dpg.generate_uuid()
                                    dpg.add_button(
                                        label="Average",
                                        callback=lambda s, a, u: signals.emit(
                                            Signal.SELECT_DATA_SETS_TO_AVERAGE,
                                        ),
                                        width=-1,
                                        tag=self.average_button,
                                    )
                                    attach_tooltip(tooltips.data_sets.average)
                        with dpg.child_window(
                            border=False, width=-2, height=-40, show=True
                        ):
                            self.data_table = DataTable()
                        with dpg.child_window(border=True, width=-2):
                            with dpg.group(horizontal=True):
                                self.toggle_points_button: int = dpg.generate_uuid()
                                dpg.add_button(
                                    label="Toggle points",
                                    tag=self.toggle_points_button,
                                    callback=lambda s, a, u: signals.emit(
                                        Signal.SELECT_DATA_POINTS_TO_TOGGLE,
                                        data=u,
                                    ),
                                )
                                attach_tooltip(tooltips.data_sets.toggle)
                                self.copy_mask_button: int = dpg.generate_uuid()
                                dpg.add_button(
                                    label="Copy mask",
                                    tag=self.copy_mask_button,
                                    callback=lambda s, a, u: signals.emit(
                                        Signal.SELECT_DATA_SET_MASK_TO_COPY,
                                        data=u,
                                    ),
                                )
                                attach_tooltip(tooltips.data_sets.copy)
                                self.subtract_impedance_button: int = (
                                    dpg.generate_uuid()
                                )
                                dpg.add_button(
                                    label="Subtract",
                                    tag=self.subtract_impedance_button,
                                    callback=lambda s, a, u: signals.emit(
                                        Signal.SELECT_IMPEDANCE_TO_SUBTRACT,
                                        data=u,
                                    ),
                                )
                                attach_tooltip(tooltips.data_sets.subtract)
                                self.enlarge_nyquist_button: int = dpg.generate_uuid()
                                self.adjust_nyquist_limits_checkbox: int = (
                                    dpg.generate_uuid()
                                )
                                dpg.add_button(
                                    label="Enlarge Nyquist",
                                    callback=self.show_enlarged_nyquist,
                                    tag=self.enlarge_nyquist_button,
                                )
                                dpg.add_checkbox(
                                    default_value=True,
                                    tag=self.adjust_nyquist_limits_checkbox,
                                )
                                attach_tooltip(tooltips.general.adjust_nyquist_limits)
                                self.enlarge_bode_button: int = dpg.generate_uuid()
                                self.adjust_bode_limits_checkbox: int = (
                                    dpg.generate_uuid()
                                )
                                dpg.add_button(
                                    label="Enlarge Bode",
                                    callback=self.show_enlarged_bode,
                                    tag=self.enlarge_bode_button,
                                )
                                dpg.add_checkbox(
                                    default_value=True,
                                    tag=self.adjust_bode_limits_checkbox,
                                )
                                attach_tooltip(tooltips.general.adjust_bode_limits)
                    self.plot_window: int = dpg.generate_uuid()
                    self.plot_width: int = 400
                    with dpg.child_window(
                        border=False,
                        width=-1,
                        no_scrollbar=True,
                        tag=self.plot_window,
                        show=True,
                    ):
                        self.nyquist_plot: Nyquist = Nyquist()
                        self.nyquist_plot.plot(
                            real=array([]),
                            imaginary=array([]),
                            label="Data",
                            line=False,
                            theme=themes.nyquist.data,
                        )
                        self.nyquist_plot.plot(
                            real=array([]),
                            imaginary=array([]),
                            label="Data",
                            line=True,
                            theme=themes.nyquist.data,
                            show_label=False,
                        )
                        self.bode_plot: Bode = Bode()
                        self.bode_plot.plot(
                            frequency=array([]),
                            magnitude=array([]),
                            phase=array([]),
                            labels=(
                                "|Z|",
                                "phi",
                            ),
                            line=False,
                            themes=(
                                themes.bode.magnitude_data,
                                themes.bode.phase_data,
                            ),
                        )
                        self.bode_plot.plot(
                            frequency=array([]),
                            magnitude=array([]),
                            phase=array([]),
                            labels=(
                                "|Z|",
                                "phi",
                            ),
                            line=True,
                            themes=(
                                themes.bode.magnitude_data,
                                themes.bode.phase_data,
                            ),
                            show_labels=False,
                        )

    def is_visible(self) -> bool:
        return dpg.is_item_visible(self.visibility_item)

    def resize(self, width: int, height: int):
        assert type(width) is int and width > 0, width
        assert type(height) is int and height > 0, height
        if not self.is_visible():
            return
        if width < (self.table_width + self.plot_width):
            if dpg.is_item_shown(self.plot_window):
                dpg.hide_item(self.plot_window)
                dpg.set_item_width(self.table_window, -1)
                dpg.set_item_label(self.enlarge_nyquist_button, "Show Nyquist")
                dpg.set_item_label(self.enlarge_bode_button, "Show Bode")
        else:
            if not dpg.is_item_shown(self.plot_window):
                dpg.show_item(self.plot_window)
                dpg.set_item_width(self.table_window, self.table_width)
                dpg.split_frame()
                dpg.set_item_label(self.enlarge_nyquist_button, "Enlarge Nyquist")
                dpg.set_item_label(self.enlarge_bode_button, "Enlarge Bode")
        if not dpg.is_item_shown(self.plot_window):
            return
        width, height = dpg.get_item_rect_size(self.plot_window)
        item: int
        for item in dpg.get_item_children(self.plot_window, slot=1):
            dpg.set_item_height(item, height / 2)

    def clear(self):
        dpg.set_value(self.data_sets_combo, "")
        dpg.set_value(self.label_input, "")
        dpg.set_value(self.path_input, "")
        dpg.set_item_user_data(self.delete_button, None)
        dpg.set_item_user_data(self.toggle_points_button, None)
        dpg.set_item_user_data(self.copy_mask_button, None)
        dpg.set_item_user_data(self.subtract_impedance_button, None)
        self.data_table.clear()
        self.nyquist_plot.clear(delete=False)
        self.bode_plot.clear(delete=False)

    def populate_data_sets(self, labels: List[str], lookup: Dict[str, DataSet]):
        assert type(labels) is list, labels
        assert type(lookup) is dict, lookup
        label: str = dpg.get_value(self.data_sets_combo) or ""
        if labels and label not in labels:
            label = labels[0]
        dpg.configure_item(
            self.data_sets_combo,
            items=labels,
            default_value=label,
            user_data=lookup,
        )

    def get_next_data_set(self) -> Optional[DataSet]:
        lookup: Dict[str, DataSet] = dpg.get_item_user_data(self.data_sets_combo)
        if not lookup:
            return None
        labels: List[str] = list(lookup.keys())
        index: int = labels.index(dpg.get_value(self.data_sets_combo)) + 1
        return lookup[labels[index % len(labels)]]

    def get_previous_data_set(self) -> Optional[DataSet]:
        lookup: Dict[str, DataSet] = dpg.get_item_user_data(self.data_sets_combo)
        if not lookup:
            return None
        labels: List[str] = list(lookup.keys())
        index: int = labels.index(dpg.get_value(self.data_sets_combo)) - 1
        return lookup[labels[index % len(labels)]]

    def select_data_set(self, data: Optional[DataSet]):
        assert type(data) is DataSet or data is None, data
        if not self.is_visible():
            self.queued_update = lambda: self.select_data_set(data)
            return
        self.queued_update = None
        if data is None:
            self.clear()
            return
        self.data_table.populate(data)
        self.data_table.update(data)
        dpg.set_value(self.data_sets_combo, data.get_label())
        dpg.configure_item(
            self.label_input,
            default_value=data.get_label(),
            user_data=data,
        )
        dpg.configure_item(
            self.path_input,
            default_value=data.get_path(),
            user_data=data,
        )
        dpg.set_item_user_data(self.delete_button, data)
        dpg.set_item_user_data(self.toggle_points_button, data)
        dpg.set_item_user_data(self.copy_mask_button, data)
        dpg.set_item_user_data(self.subtract_impedance_button, data)
        real: ndarray
        imag: ndarray
        real, imag = data.get_nyquist_data()
        self.nyquist_plot.update(
            index=0,
            real=real,
            imaginary=imag,
        )
        self.nyquist_plot.update(
            index=1,
            real=real,
            imaginary=imag,
        )
        freq: ndarray
        mag: ndarray
        phase: ndarray
        freq, mag, phase = data.get_bode_data()
        self.bode_plot.update(
            index=0,
            frequency=freq,
            magnitude=mag,
            phase=phase,
        )
        self.bode_plot.update(
            index=1,
            frequency=freq,
            magnitude=mag,
            phase=phase,
        )
        if dpg.get_value(self.adjust_nyquist_limits_checkbox):
            self.nyquist_plot.queue_limits_adjustment()
        if dpg.get_value(self.adjust_bode_limits_checkbox):
            self.bode_plot.queue_limits_adjustment()

    def show_enlarged_nyquist(self):
        signals.emit(
            Signal.SHOW_ENLARGED_PLOT,
            plot=self.nyquist_plot,
            adjust_limits=dpg.get_value(self.adjust_nyquist_limits_checkbox),
        )

    def show_enlarged_bode(self):
        signals.emit(
            Signal.SHOW_ENLARGED_PLOT,
            plot=self.bode_plot,
            adjust_limits=dpg.get_value(self.adjust_bode_limits_checkbox),
        )

    def has_active_input(self) -> bool:
        return dpg.is_item_active(self.label_input) or dpg.is_item_active(
            self.path_input
        )



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from math import floor
from typing import (
    Dict,
    List,
    Optional,
    Tuple,
)
import dearpygui.dearpygui as dpg
import matplotlib

matplotlib.use("Agg")
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.backends.backend_agg import FigureCanvasAgg
from numpy import asarray, float32
from deareis.data import (
    PlotSettings,
    Project,
)
from deareis.gui.plots import Image
from deareis.signals import Signal
from deareis.tooltips import (
    attach_tooltip,
    update_tooltip,
)
from deareis.utility import format_number
from deareis.utility import calculate_window_position_dimensions
import deareis.api.plot.mpl as mpl
import deareis.signals as signals
import deareis.tooltips as tooltips
from deareis.keybindings import (
    is_alt_down,
    is_control_down,
)
from deareis.config import Config


PREVIEW_LIMITS: Dict[str, int] = {
    "No limit": -1,
    "256 px": 2**8,
    "512 px": 2**9,
    "1024 px": 2**10,
    "2048 px": 2**11,
    "4096 px": 2**12,
    "8192 px": 2**13,
}


UNITS_PER_INCH: Dict[str, float] = {
    "Centimeters": 2.54,
    "Inches": 1.0,
}


LEGEND_LOCATIONS: Dict[str, int] = {
    "Automatic": 0,
    "Upper right": 1,
    "Upper left": 2,
    "Lower left": 3,
    "Lower right": 4,
    "Right": 5,
    "Center left": 6,
    "Center right": 7,
    "Lower center": 8,
    "Upper center": 9,
    "Center": 10,
}


EXTENSIONS: List[str] = [
    ".eps",
    ".jpg",
    ".pdf",
    ".pgf",
    ".png",
    ".ps",
    ".svg",
]


class PlotExporter:
    def __init__(self, config: Config):
        self.key_handler: int = -1
        self.settings: Optional[PlotSettings] = None
        self.project: Optional[Project] = None
        self.texture_registry: int = dpg.generate_uuid()
        self.experimental_clear_registry: bool = True
        self.experimental_disable_previews: bool = False
        dpg.add_texture_registry(tag=self.texture_registry)
        self.window: int = dpg.generate_uuid()
        with dpg.window(
            label="Export plot",
            show=False,
            modal=True,
            on_close=self.close,
            tag=self.window,
        ):
            with dpg.group(horizontal=True):
                dpg.add_spacer(width=8)
                self.sidebar: int = dpg.generate_uuid()
                with dpg.child_window(
                    width=250,
                    height=-1,
                    border=False,
                    tag=self.sidebar,
                ):
                    dpg.add_spacer(height=4)
                    with dpg.child_window(
                        border=False,
                        height=-32,
                    ):
                        self.unit_combo: int = dpg.generate_uuid()
                        self.width_input: int = dpg.generate_uuid()
                        self.height_input: int = dpg.generate_uuid()
                        self.dpi_input: int = dpg.generate_uuid()
                        self.label_padding: int = 10
                        with dpg.collapsing_header(label=" Dimensions", leaf=True):
                            with dpg.group(horizontal=True):
                                dpg.add_text("Units".rjust(self.label_padding))
                                attach_tooltip(tooltips.plotting.export_units)
                                dpg.add_combo(
                                    width=-1,
                                    items=list(UNITS_PER_INCH.keys()),
                                    default_value="",
                                    callback=lambda s, a, u: self.refresh(),
                                    tag=self.unit_combo,
                                )
                            with dpg.group(horizontal=True):
                                dpg.add_text("Width".rjust(self.label_padding))
                                attach_tooltip(tooltips.plotting.export_width)
                                dpg.add_input_float(
                                    width=-1,
                                    default_value=0.0,
                                    min_value=0.01,
                                    min_clamped=True,
                                    step=0.0,
                                    format="%.2f",
                                    on_enter=True,
                                    callback=lambda s, a, u: self.refresh(),
                                    tag=self.width_input,
                                )
                            with dpg.group(horizontal=True):
                                dpg.add_text("Height".rjust(self.label_padding))
                                attach_tooltip(tooltips.plotting.export_height)
                                dpg.add_input_float(
                                    width=-1,
                                    default_value=0.0,
                                    min_value=0.01,
                                    min_clamped=True,
                                    step=0.0,
                                    format="%.2f",
                                    on_enter=True,
                                    callback=lambda s, a, u: self.refresh(),
                                    tag=self.height_input,
                                )
                            with dpg.group(horizontal=True):
                                dpg.add_text("DPI".rjust(self.label_padding))
                                attach_tooltip(tooltips.plotting.export_dpi)
                                dpg.add_input_int(
                                    width=-1,
                                    default_value=0,
                                    min_value=1,
                                    min_clamped=True,
                                    step=0,
                                    on_enter=True,
                                    callback=lambda s, a, u: self.refresh(),
                                    tag=self.dpi_input,
                                )
                            with dpg.group(horizontal=True):
                                dpg.add_text("Result".rjust(self.label_padding))
                                attach_tooltip(tooltips.plotting.export_result)
                                tooltip_tag: int = dpg.generate_uuid()
                                self.dimensions_output: int = dpg.add_text(
                                    "", user_data=tooltip_tag
                                )
                                attach_tooltip("", tag=tooltip_tag)
                            with dpg.group(horizontal=True):
                                dpg.add_text("Preview".rjust(self.label_padding))
                                attach_tooltip(tooltips.plotting.export_preview)
                                self.preview_combo: int = dpg.generate_uuid()
                                dpg.add_combo(
                                    items=list(PREVIEW_LIMITS.keys()),
                                    default_value="",
                                    callback=lambda s, a, u: self.refresh(),
                                    tag=self.preview_combo,
                                )
                        with dpg.collapsing_header(label=" Limits", leaf=True):
                            with dpg.group(horizontal=True):
                                dpg.add_text("Min. X".rjust(self.label_padding))
                                attach_tooltip(tooltips.plotting.export_min_x)
                                self.x_min_enabled: int = dpg.generate_uuid()
                                dpg.add_checkbox(
                                    default_value=False,
                                    callback=lambda s, a, u: self.refresh(),
                                    tag=self.x_min_enabled,
                                )
                                self.x_min: int = dpg.generate_uuid()
                                dpg.add_input_float(
                                    width=-1,
                                    step=0.0,
                                    format="%.2f",
                                    on_enter=True,
                                    callback=lambda s, a, u: self.refresh(),
                                    tag=self.x_min,
                                )
                            with dpg.group(horizontal=True):
                                dpg.add_text("Max. X".rjust(self.label_padding))
                                attach_tooltip(tooltips.plotting.export_max_x)
                                self.x_max_enabled: int = dpg.generate_uuid()
                                dpg.add_checkbox(
                                    default_value=False,
                                    callback=lambda s, a, u: self.refresh(),
                                    tag=self.x_max_enabled,
                                )
                                self.x_max: int = dpg.generate_uuid()
                                dpg.add_input_float(
                                    width=-1,
                                    step=0.0,
                                    format="%.2f",
                                    on_enter=True,
                                    callback=lambda s, a, u: self.refresh(),
                                    tag=self.x_max,
                                )
                            with dpg.group(horizontal=True):
                                dpg.add_text("Min. Y".rjust(self.label_padding))
                                attach_tooltip(tooltips.plotting.export_min_y)
                                self.y_min_enabled: int = dpg.generate_uuid()
                                dpg.add_checkbox(
                                    default_value=False,
                                    callback=lambda s, a, u: self.refresh(),
                                    tag=self.y_min_enabled,
                                )
                                self.y_min: int = dpg.generate_uuid()
                                dpg.add_input_float(
                                    width=-1,
                                    step=0.0,
                                    format="%.2f",
                                    on_enter=True,
                                    callback=lambda s, a, u: self.refresh(),
                                    tag=self.y_min,
                                )
                            with dpg.group(horizontal=True):
                                dpg.add_text("Max. Y".rjust(self.label_padding))
                                attach_tooltip(tooltips.plotting.export_max_y)
                                self.y_max_enabled: int = dpg.generate_uuid()
                                dpg.add_checkbox(
                                    default_value=False,
                                    callback=lambda s, a, u: self.refresh(),
                                    tag=self.y_max_enabled,
                                )
                                self.y_max: int = dpg.generate_uuid()
                                dpg.add_input_float(
                                    width=-1,
                                    step=0.0,
                                    format="%.2f",
                                    on_enter=True,
                                    callback=lambda s, a, u: self.refresh(),
                                    tag=self.y_max,
                                )
                        with dpg.collapsing_header(label=" Layout", leaf=True):
                            with dpg.group(horizontal=True):
                                dpg.add_text("Title".rjust(self.label_padding))
                                attach_tooltip(tooltips.plotting.export_title)
                                self.show_title: int = dpg.generate_uuid()
                                dpg.add_checkbox(
                                    default_value=True,
                                    callback=lambda s, a, u: self.refresh(),
                                    tag=self.show_title,
                                )
                            with dpg.group(horizontal=True):
                                dpg.add_text("Legend".rjust(self.label_padding))
                                attach_tooltip(tooltips.plotting.export_legend)
                                self.show_legend: int = dpg.generate_uuid()
                                dpg.add_checkbox(
                                    default_value=True,
                                    callback=lambda s, a, u: self.refresh(),
                                    tag=self.show_legend,
                                )
                                self.legend_combo: int = dpg.generate_uuid()
                                dpg.add_combo(
                                    items=list(LEGEND_LOCATIONS.keys()),
                                    default_value=list(LEGEND_LOCATIONS.keys())[0],
                                    callback=lambda s, a, u: self.refresh(),
                                    width=-1,
                                    tag=self.legend_combo,
                                )
                            with dpg.group(horizontal=True):
                                dpg.add_text("Grid".rjust(self.label_padding))
                                attach_tooltip(tooltips.plotting.export_grid)
                                self.grid: int = dpg.generate_uuid()
                                dpg.add_checkbox(
                                    default_value=False,
                                    callback=lambda s, a, u: self.refresh(),
                                    tag=self.grid,
                                )
                            with dpg.group(horizontal=True):
                                dpg.add_text("Tight".rjust(self.label_padding))
                                attach_tooltip(tooltips.plotting.export_tight)
                                self.tight_layout: int = dpg.generate_uuid()
                                dpg.add_checkbox(
                                    default_value=False,
                                    callback=lambda s, a, u: self.refresh(),
                                    tag=self.tight_layout,
                                )
                        with dpg.collapsing_header(label=" Miscellaneous", leaf=True):
                            with dpg.group(horizontal=True):
                                dpg.add_text(
                                    "Points per decade".rjust(self.label_padding)
                                )
                                attach_tooltip(tooltips.plotting.export_num_per_decade)
                                self.num_per_decade: int = dpg.generate_uuid()
                                dpg.add_input_int(
                                    width=-1,
                                    default_value=0,
                                    min_value=1,
                                    min_clamped=True,
                                    step=0,
                                    on_enter=True,
                                    callback=lambda s, a, u: self.refresh(),
                                    tag=self.num_per_decade,
                                )
                    with dpg.child_window(border=False):
                        self.save_button: int = dpg.add_button(
                            label="Save as",
                            callback=lambda s, a, u: self.save(u),
                            user_data=None,
                            width=-1,
                        )
                        attach_tooltip(tooltips.plotting.export_file)
                self.image_plot: Image = Image()
        self.update_settings(config)

    def clear(self):
        fig: Optional[Figure] = dpg.get_item_user_data(self.save_button)
        if fig is not None:
            plt.close(fig)
            dpg.set_item_user_data(self.save_button, None)
        if not self.experimental_disable_previews:
            self.image_plot.clear()
            if self.experimental_clear_registry:
                dpg.delete_item(self.texture_registry, children_only=True)

    def close(self):
        if dpg.does_item_exist(self.key_handler):
            dpg.delete_item(self.key_handler)
        dpg.hide_item(self.window)
        self.clear()
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)
        # dpg.delete_item(self.key_handler)

    def update_settings(self, config: Config):
        dpg.set_value(self.unit_combo, list(UNITS_PER_INCH.keys())[config.export_units])
        dpg.set_value(self.width_input, config.export_width)
        dpg.set_value(self.height_input, config.export_height)
        dpg.set_value(self.dpi_input, config.export_dpi)
        dpg.set_value(
            self.preview_combo, list(PREVIEW_LIMITS.keys())[config.export_preview]
        )
        dpg.set_value(self.show_title, config.export_title)
        dpg.set_value(self.show_legend, config.export_legend)
        dpg.set_value(
            self.legend_combo,
            list(LEGEND_LOCATIONS.keys())[config.export_legend_location],
        )
        dpg.set_value(self.grid, config.export_grid)
        dpg.set_value(self.tight_layout, config.export_tight)
        dpg.set_value(self.num_per_decade, config.export_num_per_decade)
        self.experimental_clear_registry = config.export_experimental_clear_registry
        self.experimental_disable_previews = config.export_experimental_disable_previews

    def initialize_keybindings(self):
        self.key_handler = dpg.generate_uuid()
        with dpg.handler_registry(tag=self.key_handler):
            dpg.add_key_release_handler(
                key=dpg.mvKey_Escape,
                callback=self.close,
            )
            dpg.add_key_release_handler(
                key=dpg.mvKey_Return,
                callback=lambda: self.save(
                    dpg.get_item_user_data(self.save_button), keybinding=True
                ),
            )

    def show(self, settings: PlotSettings, project: Project):
        self.initialize_keybindings()
        self.settings = settings
        self.project = project
        dpg.show_item(self.window)
        dpg.split_frame()
        x: int
        y: int
        w: int
        h: int
        x, y, w, h = calculate_window_position_dimensions()
        dpg.configure_item(
            self.window,
            pos=(
                x,
                y,
            ),
            width=w,
            height=h,
        )
        signals.emit(Signal.BLOCK_KEYBINDINGS, window=self.window)
        self.refresh()

    def create_preview_figure(
        self, width: float, height: float, dpi: int
    ) -> Tuple[Figure, int, int]:
        assert type(width) is float and width > 0.0, width
        assert type(height) is float and height > 0.0, height
        assert type(dpi) is int and dpi > 0, dpi
        pixel_width: int = floor(width * dpi)
        pixel_height: int = floor(height * dpi)
        preview_limit: int = PREVIEW_LIMITS.get(dpg.get_value(self.preview_combo), -1)
        if preview_limit > 0 and (
            pixel_width > preview_limit or pixel_height > preview_limit
        ):
            if pixel_width >= pixel_height:
                pixel_height = floor(pixel_height / pixel_width * preview_limit)
                pixel_width = preview_limit
            else:
                pixel_width = floor(pixel_width / pixel_height * preview_limit)
                pixel_height = preview_limit
        width = pixel_width / dpi
        height = pixel_height / dpi
        fig: Figure = plt.figure(
            figsize=(
                width,
                height,
            ),
            dpi=dpi,
        )
        return (
            fig,
            pixel_width,
            pixel_height,
        )

    def create_final_figure(self, width: float, height: float, dpi: int) -> Figure:
        assert type(width) is float and width > 0.0, width
        assert type(height) is float and height > 0.0, height
        assert type(dpi) is int and dpi > 0, dpi
        pixel_width: int = floor(width * dpi)
        pixel_height: int = floor(height * dpi)
        pixel_dimensions: List[str] = [
            format_number(float(pixel_width), decimals=0, exponent=False),
            format_number(float(pixel_height), decimals=0, exponent=False),
        ]
        dpg.set_value(
            self.dimensions_output,
            f"{pixel_dimensions[0]} px * {pixel_dimensions[1]} px",
        )
        update_tooltip(
            dpg.get_item_user_data(self.dimensions_output),
            f"{pixel_dimensions[0]} px * {pixel_dimensions[1]} px",
        )
        fig: Figure = plt.figure(
            figsize=(
                width,
                height,
            ),
            dpi=dpi,
        )
        return fig

    def plot(self, fig: Figure):
        assert type(fig) is Figure, type(fig)
        assert self.settings is not None
        assert self.project is not None
        x_limits: Optional[Tuple[Optional[float], Optional[float]]] = None
        if dpg.get_value(self.x_min_enabled) or dpg.get_value(self.x_max_enabled):
            x_limits = (
                dpg.get_value(self.x_min)
                if dpg.get_value(self.x_min_enabled)
                else None,
                dpg.get_value(self.x_max)
                if dpg.get_value(self.x_max_enabled)
                else None,
            )
        y_limits: Optional[Tuple[Optional[float], Optional[float]]] = None
        if dpg.get_value(self.y_min_enabled) or dpg.get_value(self.y_max_enabled):
            y_limits = (
                dpg.get_value(self.y_min)
                if dpg.get_value(self.y_min_enabled)
                else None,
                dpg.get_value(self.y_max)
                if dpg.get_value(self.y_max_enabled)
                else None,
            )
        mpl.plot(
            self.settings,
            self.project,
            x_limits=x_limits,
            y_limits=y_limits,
            show_title=dpg.get_value(self.show_title),
            show_legend=dpg.get_value(self.show_legend),
            legend_loc=LEGEND_LOCATIONS.get(dpg.get_value(self.legend_combo), 0),
            tight_layout=dpg.get_value(self.tight_layout),
            show_grid=dpg.get_value(self.grid),
            fig=fig,
            axis=fig.gca(),
            num_per_decade=dpg.get_value(self.num_per_decade),
        )

    def refresh(self):
        width: float = dpg.get_value(self.width_input)
        height: float = dpg.get_value(self.height_input)
        dpi: int = dpg.get_value(self.dpi_input)
        upi: float = UNITS_PER_INCH.get(dpg.get_value(self.unit_combo))
        assert type(width) is float and width > 0.0, width
        assert type(height) is float and height > 0.0, height
        assert type(dpi) is int and dpi > 0, dpi
        assert type(upi) is float and upi > 0.0, upi
        self.clear()
        width = width / upi
        height = height / upi
        final_fig: Figure = self.create_final_figure(width, height, dpi)
        dpg.set_item_user_data(self.save_button, final_fig)
        self.plot(final_fig)
        # Preview
        if not self.experimental_disable_previews:
            preview_fig: Figure
            pixel_width: int
            pixel_height: int
            preview_fig, pixel_width, pixel_height = self.create_preview_figure(
                width, height, dpi
            )
            canvas: FigureCanvasAgg = FigureCanvasAgg(preview_fig)
            self.plot(preview_fig)
            canvas.draw()
            tag: int = dpg.add_raw_texture(
                pixel_width,
                pixel_height,
                asarray(canvas.buffer_rgba()).astype(float32) / 255,
                format=dpg.mvFormat_Float_rgba,
                parent=self.texture_registry,
            )
            self.image_plot.plot(
                texture=tag,
                bounds_min=(
                    0,
                    0,
                ),
                bounds_max=(
                    pixel_width,
                    pixel_height,
                ),
            )
            plt.close(preview_fig)
            self.image_plot.queue_limits_adjustment()
        x_min: float
        x_max: float
        y_min: float
        y_max: float
        x_min, x_max = final_fig.gca().get_xlim()
        y_min, y_max = final_fig.gca().get_ylim()
        if dpg.get_value(self.x_min_enabled):
            dpg.enable_item(self.x_min)
        else:
            dpg.disable_item(self.x_min)
            dpg.set_value(self.x_min, x_min)
        if dpg.get_value(self.x_max_enabled):
            dpg.enable_item(self.x_max)
        else:
            dpg.disable_item(self.x_max)
            dpg.set_value(self.x_max, x_max)
        if dpg.get_value(self.y_min_enabled):
            dpg.enable_item(self.y_min)
        else:
            dpg.disable_item(self.y_min)
            dpg.set_value(self.y_min, y_min)
        if dpg.get_value(self.y_max_enabled):
            dpg.enable_item(self.y_max)
        else:
            dpg.disable_item(self.y_max)
            dpg.set_value(self.y_max, y_max)
        if dpg.get_value(self.show_legend):
            dpg.enable_item(self.legend_combo)
        else:
            dpg.disable_item(self.legend_combo)

    def save(self, fig: Optional[Figure], keybinding: bool = False):
        if fig is None:
            return
        elif keybinding is True and not (
            is_control_down()
            if dpg.get_platform() == dpg.mvPlatform_Windows
            else is_alt_down()
        ):
            return
        self.close()
        dpg.split_frame()
        signals.emit(Signal.SAVE_PLOT, figure=fig)
